import io
import zipfile
from pathlib import Path

from .base import ModuleTestBase


class TestGithub_Workflows(ModuleTestBase):
    config_overrides = {"modules": {"github_org": {"api_key": "asdf"}}}
    modules_overrides = ["github_workflows", "github_org", "speculate"]

    data = io.BytesIO()
    with zipfile.ZipFile(data, mode="w", compression=zipfile.ZIP_DEFLATED) as zipfile:
        zipfile.writestr("test.txt", "This is some test data")
        zipfile.writestr("test2.txt", "This is some more test data")
        zipfile.writestr("folder/test3.txt", "This is yet more test data")
    data.seek(0)
    zip_content = data.getvalue()

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.github.com/zen", match_headers={"Authorization": "token asdf"}
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity",
            match_headers={"Authorization": "token asdf"},
            json={
                "login": "blacklanternsecurity",
                "id": 25311592,
                "node_id": "MDEyOk9yZ2FuaXphdGlvbjI1MzExNTky",
                "url": "https://api.github.com/orgs/blacklanternsecurity",
                "repos_url": "https://api.github.com/orgs/blacklanternsecurity/repos",
                "events_url": "https://api.github.com/orgs/blacklanternsecurity/events",
                "hooks_url": "https://api.github.com/orgs/blacklanternsecurity/hooks",
                "issues_url": "https://api.github.com/orgs/blacklanternsecurity/issues",
                "members_url": "https://api.github.com/orgs/blacklanternsecurity/members{/member}",
                "public_members_url": "https://api.github.com/orgs/blacklanternsecurity/public_members{/member}",
                "avatar_url": "https://avatars.githubusercontent.com/u/25311592?v=4",
                "description": "Security Organization",
                "name": "Black Lantern Security",
                "company": None,
                "blog": "www.blacklanternsecurity.com",
                "location": "Charleston, SC",
                "email": None,
                "twitter_username": None,
                "is_verified": False,
                "has_organization_projects": True,
                "has_repository_projects": True,
                "public_repos": 70,
                "public_gists": 0,
                "followers": 415,
                "following": 0,
                "html_url": "https://github.com/blacklanternsecurity",
                "created_at": "2017-01-24T00:14:46Z",
                "updated_at": "2022-03-28T11:39:03Z",
                "archived_at": None,
                "type": "Organization",
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/orgs/blacklanternsecurity/repos?per_page=100&page=1",
            match_headers={"Authorization": "token asdf"},
            json=[
                {
                    "id": 459780477,
                    "node_id": "R_kgDOG2exfQ",
                    "name": "test_keys",
                    "full_name": "blacklanternsecurity/test_keys",
                    "private": False,
                    "owner": {
                        "login": "blacklanternsecurity",
                        "id": 79229934,
                        "node_id": "MDEyOk9yZ2FuaXphdGlvbjc5MjI5OTM0",
                        "avatar_url": "https://avatars.githubusercontent.com/u/79229934?v=4",
                        "gravatar_id": "",
                        "url": "https://api.github.com/users/blacklanternsecurity",
                        "html_url": "https://github.com/blacklanternsecurity",
                        "followers_url": "https://api.github.com/users/blacklanternsecurity/followers",
                        "following_url": "https://api.github.com/users/blacklanternsecurity/following{/other_user}",
                        "gists_url": "https://api.github.com/users/blacklanternsecurity/gists{/gist_id}",
                        "starred_url": "https://api.github.com/users/blacklanternsecurity/starred{/owner}{/repo}",
                        "subscriptions_url": "https://api.github.com/users/blacklanternsecurity/subscriptions",
                        "organizations_url": "https://api.github.com/users/blacklanternsecurity/orgs",
                        "repos_url": "https://api.github.com/users/blacklanternsecurity/repos",
                        "events_url": "https://api.github.com/users/blacklanternsecurity/events{/privacy}",
                        "received_events_url": "https://api.github.com/users/blacklanternsecurity/received_events",
                        "type": "Organization",
                        "site_admin": False,
                    },
                    "html_url": "https://github.com/blacklanternsecurity/bbot",
                    "description": None,
                    "fork": False,
                    "url": "https://api.github.com/repos/blacklanternsecurity/test_keys",
                    "forks_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/forks",
                    "keys_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/keys{/key_id}",
                    "collaborators_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/collaborators{/collaborator}",
                    "teams_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/teams",
                    "hooks_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/hooks",
                    "issue_events_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues/events{/number}",
                    "events_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/events",
                    "assignees_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/assignees{/user}",
                    "branches_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/branches{/branch}",
                    "tags_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/tags",
                    "blobs_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/blobs{/sha}",
                    "git_tags_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/tags{/sha}",
                    "git_refs_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/refs{/sha}",
                    "trees_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/trees{/sha}",
                    "statuses_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/statuses/{sha}",
                    "languages_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/languages",
                    "stargazers_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/stargazers",
                    "contributors_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/contributors",
                    "subscribers_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/subscribers",
                    "subscription_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/subscription",
                    "commits_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/commits{/sha}",
                    "git_commits_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/git/commits{/sha}",
                    "comments_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/comments{/number}",
                    "issue_comment_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues/comments{/number}",
                    "contents_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/contents/{+path}",
                    "compare_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/compare/{base}...{head}",
                    "merges_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/merges",
                    "archive_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/{archive_format}{/ref}",
                    "downloads_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/downloads",
                    "issues_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/issues{/number}",
                    "pulls_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/pulls{/number}",
                    "milestones_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/milestones{/number}",
                    "notifications_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/notifications{?since,all,participating}",
                    "labels_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/labels{/name}",
                    "releases_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/releases{/id}",
                    "deployments_url": "https://api.github.com/repos/blacklanternsecurity/test_keys/deployments",
                    "created_at": "2022-02-15T23:10:51Z",
                    "updated_at": "2023-09-02T12:20:13Z",
                    "pushed_at": "2023-10-19T02:56:46Z",
                    "git_url": "git://github.com/blacklanternsecurity/test_keys.git",
                    "ssh_url": "git@github.com:blacklanternsecurity/test_keys.git",
                    "clone_url": "https://github.com/blacklanternsecurity/bbot.git",
                    "svn_url": "https://github.com/blacklanternsecurity/bbot",
                    "homepage": None,
                    "size": 2,
                    "stargazers_count": 2,
                    "watchers_count": 2,
                    "language": None,
                    "has_issues": True,
                    "has_projects": True,
                    "has_downloads": True,
                    "has_wiki": True,
                    "has_pages": False,
                    "has_discussions": False,
                    "forks_count": 32,
                    "mirror_url": None,
                    "archived": False,
                    "disabled": False,
                    "open_issues_count": 2,
                    "license": None,
                    "allow_forking": True,
                    "is_template": False,
                    "web_commit_signoff_required": False,
                    "topics": [],
                    "visibility": "public",
                    "forks": 32,
                    "open_issues": 2,
                    "watchers": 2,
                    "default_branch": "main",
                    "permissions": {"admin": False, "maintain": False, "push": False, "triage": False, "pull": True},
                }
            ],
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/repos/blacklanternsecurity/bbot/actions/workflows?per_page=100&page=1",
            match_headers={"Authorization": "token asdf"},
            json={
                "total_count": 3,
                "workflows": [
                    {
                        "id": 22452226,
                        "node_id": "W_kwDOG_O3ns4BVpgC",
                        "name": "tests",
                        "path": ".github/workflows/tests.yml",
                        "state": "active",
                        "created_at": "2022-03-23T15:09:22.000Z",
                        "updated_at": "2022-09-27T17:49:34.000Z",
                        "url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/workflows/22452226",
                        "html_url": "https://github.com/blacklanternsecurity/bbot/blob/stable/.github/workflows/tests.yml",
                        "badge_url": "https://github.com/blacklanternsecurity/bbot/workflows/tests/badge.svg",
                    },
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/repos/blacklanternsecurity/bbot/actions/workflows/22452226/runs?status=success&per_page=1",
            match_headers={"Authorization": "token asdf"},
            json={
                "total_count": 2993,
                "workflow_runs": [
                    {
                        "id": 8839360698,
                        "name": "tests",
                        "node_id": "WFR_kwLOG_O3ns8AAAACDt3wug",
                        "head_branch": "dnsbrute-helperify",
                        "head_sha": "c5de1360e8e5ccba04b23035f675a529282b7dc2",
                        "path": ".github/workflows/tests.yml",
                        "display_title": "Helperify Massdns",
                        "run_number": 4520,
                        "event": "pull_request",
                        "status": "completed",
                        "conclusion": "success",
                        "workflow_id": 22452226,
                        "check_suite_id": 23162098295,
                        "check_suite_node_id": "CS_kwDOG_O3ns8AAAAFZJGSdw",
                        "url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698",
                        "html_url": "https://github.com/blacklanternsecurity/bbot/actions/runs/8839360698",
                        "pull_requests": [
                            {
                                "url": "https://api.github.com/repos/blacklanternsecurity/bbot/pulls/1303",
                                "id": 1839332952,
                                "number": 1303,
                                "head": {
                                    "ref": "dnsbrute-helperify",
                                    "sha": "c5de1360e8e5ccba04b23035f675a529282b7dc2",
                                    "repo": {
                                        "id": 468957086,
                                        "url": "https://api.github.com/repos/blacklanternsecurity/bbot",
                                        "name": "bbot",
                                    },
                                },
                                "base": {
                                    "ref": "faster-regexes",
                                    "sha": "7baf219c7f3a4ba165639c5ddb62322453a8aea8",
                                    "repo": {
                                        "id": 468957086,
                                        "url": "https://api.github.com/repos/blacklanternsecurity/bbot",
                                        "name": "bbot",
                                    },
                                },
                            }
                        ],
                        "created_at": "2024-04-25T21:04:32Z",
                        "updated_at": "2024-04-25T21:19:43Z",
                        "actor": {
                            "login": "TheTechromancer",
                            "id": 20261699,
                            "node_id": "MDQ6VXNlcjIwMjYxNjk5",
                            "avatar_url": "https://avatars.githubusercontent.com/u/20261699?v=4",
                            "gravatar_id": "",
                            "url": "https://api.github.com/users/TheTechromancer",
                            "html_url": "https://github.com/TheTechromancer",
                            "followers_url": "https://api.github.com/users/TheTechromancer/followers",
                            "following_url": "https://api.github.com/users/TheTechromancer/following{/other_user}",
                            "gists_url": "https://api.github.com/users/TheTechromancer/gists{/gist_id}",
                            "starred_url": "https://api.github.com/users/TheTechromancer/starred{/owner}{/repo}",
                            "subscriptions_url": "https://api.github.com/users/TheTechromancer/subscriptions",
                            "organizations_url": "https://api.github.com/users/TheTechromancer/orgs",
                            "repos_url": "https://api.github.com/users/TheTechromancer/repos",
                            "events_url": "https://api.github.com/users/TheTechromancer/events{/privacy}",
                            "received_events_url": "https://api.github.com/users/TheTechromancer/received_events",
                            "type": "User",
                            "site_admin": False,
                        },
                        "run_attempt": 1,
                        "referenced_workflows": [],
                        "run_started_at": "2024-04-25T21:04:32Z",
                        "triggering_actor": {
                            "login": "TheTechromancer",
                            "id": 20261699,
                            "node_id": "MDQ6VXNlcjIwMjYxNjk5",
                            "avatar_url": "https://avatars.githubusercontent.com/u/20261699?v=4",
                            "gravatar_id": "",
                            "url": "https://api.github.com/users/TheTechromancer",
                            "html_url": "https://github.com/TheTechromancer",
                            "followers_url": "https://api.github.com/users/TheTechromancer/followers",
                            "following_url": "https://api.github.com/users/TheTechromancer/following{/other_user}",
                            "gists_url": "https://api.github.com/users/TheTechromancer/gists{/gist_id}",
                            "starred_url": "https://api.github.com/users/TheTechromancer/starred{/owner}{/repo}",
                            "subscriptions_url": "https://api.github.com/users/TheTechromancer/subscriptions",
                            "organizations_url": "https://api.github.com/users/TheTechromancer/orgs",
                            "repos_url": "https://api.github.com/users/TheTechromancer/repos",
                            "events_url": "https://api.github.com/users/TheTechromancer/events{/privacy}",
                            "received_events_url": "https://api.github.com/users/TheTechromancer/received_events",
                            "type": "User",
                            "site_admin": False,
                        },
                        "jobs_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/jobs",
                        "logs_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/logs",
                        "check_suite_url": "https://api.github.com/repos/blacklanternsecurity/bbot/check-suites/23162098295",
                        "artifacts_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/artifacts",
                        "cancel_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/cancel",
                        "rerun_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/rerun",
                        "previous_attempt_url": None,
                        "workflow_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/workflows/22452226",
                        "head_commit": {
                            "id": "c5de1360e8e5ccba04b23035f675a529282b7dc2",
                            "tree_id": "fe9b345c0745a5bbacb806225e92e1c48fccf35c",
                            "message": "remove debug message",
                            "timestamp": "2024-04-25T21:02:37Z",
                            "author": {"name": "TheTechromancer", "email": "thetechromancer@protonmail.com"},
                            "committer": {"name": "TheTechromancer", "email": "thetechromancer@protonmail.com"},
                        },
                        "repository": {
                            "id": 468957086,
                            "node_id": "R_kgDOG_O3ng",
                            "name": "bbot",
                            "full_name": "blacklanternsecurity/bbot",
                            "private": False,
                            "owner": {
                                "login": "blacklanternsecurity",
                                "id": 25311592,
                                "node_id": "MDEyOk9yZ2FuaXphdGlvbjI1MzExNTky",
                                "avatar_url": "https://avatars.githubusercontent.com/u/25311592?v=4",
                                "gravatar_id": "",
                                "url": "https://api.github.com/users/blacklanternsecurity",
                                "html_url": "https://github.com/blacklanternsecurity",
                                "followers_url": "https://api.github.com/users/blacklanternsecurity/followers",
                                "following_url": "https://api.github.com/users/blacklanternsecurity/following{/other_user}",
                                "gists_url": "https://api.github.com/users/blacklanternsecurity/gists{/gist_id}",
                                "starred_url": "https://api.github.com/users/blacklanternsecurity/starred{/owner}{/repo}",
                                "subscriptions_url": "https://api.github.com/users/blacklanternsecurity/subscriptions",
                                "organizations_url": "https://api.github.com/users/blacklanternsecurity/orgs",
                                "repos_url": "https://api.github.com/users/blacklanternsecurity/repos",
                                "events_url": "https://api.github.com/users/blacklanternsecurity/events{/privacy}",
                                "received_events_url": "https://api.github.com/users/blacklanternsecurity/received_events",
                                "type": "Organization",
                                "site_admin": False,
                            },
                            "html_url": "https://github.com/blacklanternsecurity/bbot",
                            "description": "A recursive internet scanner for hackers.",
                            "fork": False,
                            "url": "https://api.github.com/repos/blacklanternsecurity/bbot",
                            "forks_url": "https://api.github.com/repos/blacklanternsecurity/bbot/forks",
                            "keys_url": "https://api.github.com/repos/blacklanternsecurity/bbot/keys{/key_id}",
                            "collaborators_url": "https://api.github.com/repos/blacklanternsecurity/bbot/collaborators{/collaborator}",
                            "teams_url": "https://api.github.com/repos/blacklanternsecurity/bbot/teams",
                            "hooks_url": "https://api.github.com/repos/blacklanternsecurity/bbot/hooks",
                            "issue_events_url": "https://api.github.com/repos/blacklanternsecurity/bbot/issues/events{/number}",
                            "events_url": "https://api.github.com/repos/blacklanternsecurity/bbot/events",
                            "assignees_url": "https://api.github.com/repos/blacklanternsecurity/bbot/assignees{/user}",
                            "branches_url": "https://api.github.com/repos/blacklanternsecurity/bbot/branches{/branch}",
                            "tags_url": "https://api.github.com/repos/blacklanternsecurity/bbot/tags",
                            "blobs_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/blobs{/sha}",
                            "git_tags_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/tags{/sha}",
                            "git_refs_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/refs{/sha}",
                            "trees_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/trees{/sha}",
                            "statuses_url": "https://api.github.com/repos/blacklanternsecurity/bbot/statuses/{sha}",
                            "languages_url": "https://api.github.com/repos/blacklanternsecurity/bbot/languages",
                            "stargazers_url": "https://api.github.com/repos/blacklanternsecurity/bbot/stargazers",
                            "contributors_url": "https://api.github.com/repos/blacklanternsecurity/bbot/contributors",
                            "subscribers_url": "https://api.github.com/repos/blacklanternsecurity/bbot/subscribers",
                            "subscription_url": "https://api.github.com/repos/blacklanternsecurity/bbot/subscription",
                            "commits_url": "https://api.github.com/repos/blacklanternsecurity/bbot/commits{/sha}",
                            "git_commits_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/commits{/sha}",
                            "comments_url": "https://api.github.com/repos/blacklanternsecurity/bbot/comments{/number}",
                            "issue_comment_url": "https://api.github.com/repos/blacklanternsecurity/bbot/issues/comments{/number}",
                            "contents_url": "https://api.github.com/repos/blacklanternsecurity/bbot/contents/{+path}",
                            "compare_url": "https://api.github.com/repos/blacklanternsecurity/bbot/compare/{base}...{head}",
                            "merges_url": "https://api.github.com/repos/blacklanternsecurity/bbot/merges",
                            "archive_url": "https://api.github.com/repos/blacklanternsecurity/bbot/{archive_format}{/ref}",
                            "downloads_url": "https://api.github.com/repos/blacklanternsecurity/bbot/downloads",
                            "issues_url": "https://api.github.com/repos/blacklanternsecurity/bbot/issues{/number}",
                            "pulls_url": "https://api.github.com/repos/blacklanternsecurity/bbot/pulls{/number}",
                            "milestones_url": "https://api.github.com/repos/blacklanternsecurity/bbot/milestones{/number}",
                            "notifications_url": "https://api.github.com/repos/blacklanternsecurity/bbot/notifications{?since,all,participating}",
                            "labels_url": "https://api.github.com/repos/blacklanternsecurity/bbot/labels{/name}",
                            "releases_url": "https://api.github.com/repos/blacklanternsecurity/bbot/releases{/id}",
                            "deployments_url": "https://api.github.com/repos/blacklanternsecurity/bbot/deployments",
                        },
                        "head_repository": {
                            "id": 468957086,
                            "node_id": "R_kgDOG_O3ng",
                            "name": "bbot",
                            "full_name": "blacklanternsecurity/bbot",
                            "private": False,
                            "owner": {
                                "login": "blacklanternsecurity",
                                "id": 25311592,
                                "node_id": "MDEyOk9yZ2FuaXphdGlvbjI1MzExNTky",
                                "avatar_url": "https://avatars.githubusercontent.com/u/25311592?v=4",
                                "gravatar_id": "",
                                "url": "https://api.github.com/users/blacklanternsecurity",
                                "html_url": "https://github.com/blacklanternsecurity",
                                "followers_url": "https://api.github.com/users/blacklanternsecurity/followers",
                                "following_url": "https://api.github.com/users/blacklanternsecurity/following{/other_user}",
                                "gists_url": "https://api.github.com/users/blacklanternsecurity/gists{/gist_id}",
                                "starred_url": "https://api.github.com/users/blacklanternsecurity/starred{/owner}{/repo}",
                                "subscriptions_url": "https://api.github.com/users/blacklanternsecurity/subscriptions",
                                "organizations_url": "https://api.github.com/users/blacklanternsecurity/orgs",
                                "repos_url": "https://api.github.com/users/blacklanternsecurity/repos",
                                "events_url": "https://api.github.com/users/blacklanternsecurity/events{/privacy}",
                                "received_events_url": "https://api.github.com/users/blacklanternsecurity/received_events",
                                "type": "Organization",
                                "site_admin": False,
                            },
                            "html_url": "https://github.com/blacklanternsecurity/bbot",
                            "description": "A recursive internet scanner for hackers.",
                            "fork": False,
                            "url": "https://api.github.com/repos/blacklanternsecurity/bbot",
                            "forks_url": "https://api.github.com/repos/blacklanternsecurity/bbot/forks",
                            "keys_url": "https://api.github.com/repos/blacklanternsecurity/bbot/keys{/key_id}",
                            "collaborators_url": "https://api.github.com/repos/blacklanternsecurity/bbot/collaborators{/collaborator}",
                            "teams_url": "https://api.github.com/repos/blacklanternsecurity/bbot/teams",
                            "hooks_url": "https://api.github.com/repos/blacklanternsecurity/bbot/hooks",
                            "issue_events_url": "https://api.github.com/repos/blacklanternsecurity/bbot/issues/events{/number}",
                            "events_url": "https://api.github.com/repos/blacklanternsecurity/bbot/events",
                            "assignees_url": "https://api.github.com/repos/blacklanternsecurity/bbot/assignees{/user}",
                            "branches_url": "https://api.github.com/repos/blacklanternsecurity/bbot/branches{/branch}",
                            "tags_url": "https://api.github.com/repos/blacklanternsecurity/bbot/tags",
                            "blobs_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/blobs{/sha}",
                            "git_tags_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/tags{/sha}",
                            "git_refs_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/refs{/sha}",
                            "trees_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/trees{/sha}",
                            "statuses_url": "https://api.github.com/repos/blacklanternsecurity/bbot/statuses/{sha}",
                            "languages_url": "https://api.github.com/repos/blacklanternsecurity/bbot/languages",
                            "stargazers_url": "https://api.github.com/repos/blacklanternsecurity/bbot/stargazers",
                            "contributors_url": "https://api.github.com/repos/blacklanternsecurity/bbot/contributors",
                            "subscribers_url": "https://api.github.com/repos/blacklanternsecurity/bbot/subscribers",
                            "subscription_url": "https://api.github.com/repos/blacklanternsecurity/bbot/subscription",
                            "commits_url": "https://api.github.com/repos/blacklanternsecurity/bbot/commits{/sha}",
                            "git_commits_url": "https://api.github.com/repos/blacklanternsecurity/bbot/git/commits{/sha}",
                            "comments_url": "https://api.github.com/repos/blacklanternsecurity/bbot/comments{/number}",
                            "issue_comment_url": "https://api.github.com/repos/blacklanternsecurity/bbot/issues/comments{/number}",
                            "contents_url": "https://api.github.com/repos/blacklanternsecurity/bbot/contents/{+path}",
                            "compare_url": "https://api.github.com/repos/blacklanternsecurity/bbot/compare/{base}...{head}",
                            "merges_url": "https://api.github.com/repos/blacklanternsecurity/bbot/merges",
                            "archive_url": "https://api.github.com/repos/blacklanternsecurity/bbot/{archive_format}{/ref}",
                            "downloads_url": "https://api.github.com/repos/blacklanternsecurity/bbot/downloads",
                            "issues_url": "https://api.github.com/repos/blacklanternsecurity/bbot/issues{/number}",
                            "pulls_url": "https://api.github.com/repos/blacklanternsecurity/bbot/pulls{/number}",
                            "milestones_url": "https://api.github.com/repos/blacklanternsecurity/bbot/milestones{/number}",
                            "notifications_url": "https://api.github.com/repos/blacklanternsecurity/bbot/notifications{?since,all,participating}",
                            "labels_url": "https://api.github.com/repos/blacklanternsecurity/bbot/labels{/name}",
                            "releases_url": "https://api.github.com/repos/blacklanternsecurity/bbot/releases{/id}",
                            "deployments_url": "https://api.github.com/repos/blacklanternsecurity/bbot/deployments",
                        },
                    },
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/logs",
            match_headers={"Authorization": "token asdf"},
            headers={
                "location": "https://productionresultssa10.blob.core.windows.net/actions-results/7beb304e-f42c-4830-a027-4f5dec53107d/workflow-job-run-3a559e2a-952e-58d2-b8db-2e604a9266d7/logs/steps/step-logs-0e34a19a-18b0-4208-b27a-f8c031db2d17.txt?rsct=text%2Fplain&se=2024-04-26T16%3A25%3A39Z&sig=a%2FiN8dOw0e3tiBQZAfr80veI8OYChb9edJ1eFY136B4%3D&sp=r&spr=https&sr=b&st=2024-04-26T16%3A15%3A34Z&sv=2021-12-02"
            },
            status_code=302,
        )
        module_test.httpx_mock.add_response(
            url="https://productionresultssa10.blob.core.windows.net/actions-results/7beb304e-f42c-4830-a027-4f5dec53107d/workflow-job-run-3a559e2a-952e-58d2-b8db-2e604a9266d7/logs/steps/step-logs-0e34a19a-18b0-4208-b27a-f8c031db2d17.txt?rsct=text%2Fplain&se=2024-04-26T16%3A25%3A39Z&sig=a%2FiN8dOw0e3tiBQZAfr80veI8OYChb9edJ1eFY136B4%3D&sp=r&spr=https&sr=b&st=2024-04-26T16%3A15%3A34Z&sv=2021-12-02",
            content=self.zip_content,
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/repos/blacklanternsecurity/bbot/actions/runs/8839360698/artifacts",
            match_headers={"Authorization": "token asdf"},
            json={
                "total_count": 1,
                "artifacts": [
                    {
                        "id": 1829832535,
                        "node_id": "MDg6QXJ0aWZhY3QxODI5ODMyNTM1",
                        "name": "build.tar.gz",
                        "size_in_bytes": 245770648,
                        "url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/artifacts/1829832535",
                        "archive_download_url": "https://api.github.com/repos/blacklanternsecurity/bbot/actions/artifacts/1829832535/zip",
                        "expired": False,
                        "created_at": "2024-08-19T22:32:17Z",
                        "updated_at": "2024-08-19T22:32:18Z",
                        "expires_at": "2024-09-02T22:21:59Z",
                        "workflow_run": {
                            "id": 10461468466,
                            "repository_id": 89290483,
                            "head_repository_id": 799444840,
                            "head_branch": "not-a-real-branch",
                            "head_sha": "1eeb5354ab7b1e4141b8a6473846e2a5ea0dd2c6",
                        },
                    }
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.github.com/repos/blacklanternsecurity/bbot/actions/artifacts/1829832535/zip",
            match_headers={"Authorization": "token asdf"},
            headers={
                "location": "https://pipelinesghubeus22.actions.githubusercontent.com/uYHz4cw2WwYcB2EU57uoCs3MaEDiz8veiVlAtReP3xevBriD1h/_apis/pipelines/1/runs/214601/signedartifactscontent?artifactName=build.tar.gz&urlExpires=2024-08-20T14%3A41%3A41.8000556Z&urlSigningMethod=HMACV2&urlSignature=OOBxLx4eE5A8uHjxOIvQtn3cLFQOBW927mg0hcTHO6U%3D"
            },
            status_code=302,
        )
        module_test.httpx_mock.add_response(
            url="https://pipelinesghubeus22.actions.githubusercontent.com/uYHz4cw2WwYcB2EU57uoCs3MaEDiz8veiVlAtReP3xevBriD1h/_apis/pipelines/1/runs/214601/signedartifactscontent?artifactName=build.tar.gz&urlExpires=2024-08-20T14%3A41%3A41.8000556Z&urlSigningMethod=HMACV2&urlSignature=OOBxLx4eE5A8uHjxOIvQtn3cLFQOBW927mg0hcTHO6U%3D",
            content=self.zip_content,
        )

    def check(self, module_test, events):
        assert len(events) == 9
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "blacklanternsecurity.com" and e.scope_distance == 0
            ]
        ), "Failed to emit target DNS_NAME"
        assert 1 == len(
            [e for e in events if e.type == "ORG_STUB" and e.data == "blacklanternsecurity" and e.scope_distance == 0]
        ), "Failed to find ORG_STUB"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "blacklanternsecurity"
                and e.data["url"] == "https://github.com/blacklanternsecurity"
                and str(e.module) == "github_org"
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity github"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and "git" in e.tags
                and e.data["url"] == "https://github.com/blacklanternsecurity/bbot"
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity github repo"
        filesystem_events = [e for e in events if e.type == "FILESYSTEM"]
        assert 3 == len(filesystem_events), filesystem_events
        for filesystem_event in filesystem_events:
            file = Path(filesystem_event.data["path"])
            assert file.is_file(), "Destination file does not exist"



import traceback

from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class censys(subdomain_enum_apikey):
    """
    Base template for Censys API modules.
    Provides common authentication and API request handling.
    """

    options = {"api_key": ""}
    options_desc = {"api_key": "Censys.io API Key in the format of 'key:secret'"}

    base_url = "https://search.censys.io/api"

    async def setup(self):
        await super().setup()
        api_keys = set()
        for module_name in ("censys", "censys_dns", "censys_ip"):
            module_config = self.scan.config.get("modules", {}).get(module_name, {})
            api_key = module_config.get("api_key", "")
            if isinstance(api_key, str):
                api_key = [api_key]
            for key in api_key:
                key = key.strip()
                if key:
                    api_keys.add(key)
        if not api_keys:
            if self.auth_required:
                return None, "No API key set"
        self.api_key = api_keys.pop() if api_keys else ""
        try:
            await self.ping()
            self.hugesuccess("API is ready")
            return True
        except Exception as e:
            self.trace(traceback.format_exc())
            return None, f"Error with API ({str(e).strip()})"

    async def ping(self):
        url = f"{self.base_url}/v1/account"
        resp = await self.api_request(url, retry_on_http_429=False)
        d = resp.json()
        assert isinstance(d, dict), f"Invalid response from {url}: {resp}"
        quota = d.get("quota", {})
        used = int(quota.get("used", 0))
        allowance = int(quota.get("allowance", 0))
        assert used < allowance, "No quota remaining"

    def prepare_api_request(self, url, kwargs):
        api_id, api_secret = self.api_key.split(":", 1)
        kwargs["auth"] = (api_id, api_secret)
        return url, kwargs



from bbot.modules.templates.subdomain_enum import subdomain_enum


class pgp(subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS"]
    flags = ["passive", "email-enum", "safe"]
    meta = {
        "description": "Query common PGP servers for email addresses",
        "created_date": "2022-08-10",
        "author": "@TheTechromancer",
    }
    # TODO: scan for Web Key Directory (/.well-known/openpgpkey/)
    options = {
        "search_urls": [
            "https://keyserver.ubuntu.com/pks/lookup?fingerprint=on&op=vindex&search=<query>",
            "http://the.earth.li:11371/pks/lookup?fingerprint=on&op=vindex&search=<query>",
            "https://pgpkeys.eu/pks/lookup?search=<query>&op=index",
            "https://pgp.mit.edu/pks/lookup?search=<query>&op=index",
        ]
    }
    options_desc = {"search_urls": "PGP key servers to search"}

    async def handle_event(self, event):
        query = self.make_query(event)
        results = await self.query(query)
        if results:
            for email, keyserver in results:
                await self.emit_event(
                    email,
                    "EMAIL_ADDRESS",
                    event,
                    abort_if=self.abort_if,
                    context=f'{{module}} queried PGP keyserver {keyserver} for "{query}" and found {{event.type}}: {{event.data}}',
                )

    async def query(self, query):
        results = set()
        urls = self.config.get("search_urls", [])
        urls = [url.replace("<query>", self.helpers.quote(query)) for url in urls]
        async for url, response in self.helpers.request_batch(urls):
            keyserver = self.helpers.urlparse(url).netloc
            response = await self.helpers.request(url)
            if response is not None:
                for email in await self.helpers.re.extract_emails(response.text):
                    email = email.lower()
                    if email.endswith(query):
                        results.add((email, keyserver))
        return results



from .base import ModuleTestBase


class TestSpeculate_Subdirectories(ModuleTestBase):
    targets = ["http://127.0.0.1:8888/subdir1/subdir2/"]
    modules_overrides = ["httpx", "speculate"]

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/subdir1/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        expect_args = {"method": "GET", "uri": "/subdir1/subdir2/"}
        respond_args = {"response_data": "alive"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert any(e.type == "URL_UNVERIFIED" and e.data == "http://127.0.0.1:8888/subdir1/" for e in events)


class TestSpeculate_OpenPorts(ModuleTestBase):
    targets = ["evilcorp.com"]
    modules_overrides = ["speculate", "certspotter", "shodan_idb"]
    config_overrides = {"speculate": True}

    async def setup_before_prep(self, module_test):
        await module_test.mock_dns(
            {
                "evilcorp.com": {"A": ["127.0.254.1"]},
                "asdf.evilcorp.com": {"A": ["127.0.254.2"]},
            }
        )

        module_test.httpx_mock.add_response(
            url="https://api.certspotter.com/v1/issuances?domain=evilcorp.com&include_subdomains=true&expand=dns_names",
            json=[{"dns_names": ["*.asdf.evilcorp.com"]}],
        )

        from bbot.modules.base import BaseModule

        class DummyModule(BaseModule):
            _name = "dummy"
            watched_events = ["OPEN_TCP_PORT"]
            scope_distance_modifier = 10
            accept_dupes = True

            async def setup(self):
                self.events = []
                return True

            async def handle_event(self, event):
                self.events.append(event)

        module_test.scan.modules["dummy"] = DummyModule(module_test.scan)

    def check(self, module_test, events):
        events_data = set()
        for e in module_test.scan.modules["dummy"].events:
            events_data.add(e.data)
        assert all(
            x in events_data
            for x in ("evilcorp.com:80", "evilcorp.com:443", "asdf.evilcorp.com:80", "asdf.evilcorp.com:443")
        )


class TestSpeculate_OpenPorts_Portscanner(TestSpeculate_OpenPorts):
    targets = ["evilcorp.com"]
    modules_overrides = ["speculate", "certspotter", "portscan"]
    config_overrides = {"speculate": True}

    def check(self, module_test, events):
        events_data = set()
        for e in module_test.scan.modules["dummy"].events:
            events_data.add(e.data)
        assert not any(
            x in events_data
            for x in ("evilcorp.com:80", "evilcorp.com:443", "asdf.evilcorp.com:80", "asdf.evilcorp.com:443")
        )



DEP_FFUF = [
    {
        "name": "Download ffuf",
        "unarchive": {
            "src": "https://github.com/ffuf/ffuf/releases/download/v#{BBOT_DEPS_FFUF_VERSION}/ffuf_#{BBOT_DEPS_FFUF_VERSION}_#{BBOT_OS}_#{BBOT_CPU_ARCH_GOLANG}.tar.gz",
            "include": "ffuf",
            "dest": "#{BBOT_TOOLS}",
            "remote_src": True,
        },
    }
]

DEP_DOCKER = [
    {
        "name": "Check if Docker is already installed",
        "command": "docker --version",
        "register": "docker_installed",
        "ignore_errors": True,
    },
    {
        "name": "Install Docker (Non-Debian)",
        "package": {"name": "docker", "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] != 'Debian' and docker_installed.rc != 0",
    },
    {
        "name": "Install Docker (Debian)",
        "package": {
            "name": "docker.io",
            "state": "present",
        },
        "become": True,
        "when": "ansible_facts['os_family'] == 'Debian' and docker_installed.rc != 0",
    },
]

DEP_MASSDNS = [
    {
        "name": "install dev tools",
        "package": {"name": ["gcc", "git", "make"], "state": "present"},
        "become": True,
        "ignore_errors": True,
    },
    {
        "name": "Download massdns source code",
        "git": {
            "repo": "https://github.com/blechschmidt/massdns.git",
            "dest": "#{BBOT_TEMP}/massdns",
            "single_branch": True,
            "version": "master",
        },
    },
    {
        "name": "Build massdns (Linux)",
        "command": {"chdir": "#{BBOT_TEMP}/massdns", "cmd": "make", "creates": "#{BBOT_TEMP}/massdns/bin/massdns"},
        "when": "ansible_facts['system'] == 'Linux'",
    },
    {
        "name": "Build massdns (non-Linux)",
        "command": {
            "chdir": "#{BBOT_TEMP}/massdns",
            "cmd": "make nolinux",
            "creates": "#{BBOT_TEMP}/massdns/bin/massdns",
        },
        "when": "ansible_facts['system'] != 'Linux'",
    },
    {
        "name": "Install massdns",
        "copy": {"src": "#{BBOT_TEMP}/massdns/bin/massdns", "dest": "#{BBOT_TOOLS}/", "mode": "u+x,g+x,o+x"},
    },
]

DEP_CHROMIUM = [
    {
        "name": "Install Chromium (Non-Debian)",
        "package": {"name": "chromium", "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] != 'Debian'",
        "ignore_errors": True,
    },
    {
        "name": "Install Chromium dependencies (Ubuntu 24.04)",
        "package": {
            "name": "libasound2t64,libatk-bridge2.0-0,libatk1.0-0,libcairo2,libcups2,libdrm2,libgbm1,libnss3,libpango-1.0-0,libglib2.0-0,libxcomposite1,libxdamage1,libxfixes3,libxkbcommon0,libxrandr2",
            "state": "present",
        },
        "become": True,
        "when": "ansible_facts['distribution'] == 'Ubuntu' and ansible_facts['distribution_version'] == '24.04'",
        "ignore_errors": True,
    },
    {
        "name": "Install Chromium dependencies (Other Debian-based)",
        "package": {
            "name": "libasound2,libatk-bridge2.0-0,libatk1.0-0,libcairo2,libcups2,libdrm2,libgbm1,libnss3,libpango-1.0-0,libglib2.0-0,libxcomposite1,libxdamage1,libxfixes3,libxkbcommon0,libxrandr2",
            "state": "present",
        },
        "become": True,
        "when": "ansible_facts['os_family'] == 'Debian' and not (ansible_facts['distribution'] == 'Ubuntu' and ansible_facts['distribution_version'] == '24.04')",
        "ignore_errors": True,
    },
    {
        "name": "Get latest Chromium version (Debian)",
        "uri": {
            "url": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Linux_x64%2FLAST_CHANGE?alt=media",
            "return_content": True,
        },
        "register": "chromium_version",
        "when": "ansible_facts['os_family'] == 'Debian'",
        "ignore_errors": True,
    },
    {
        "name": "Get latest Chromium version (Darwin x86_64)",
        "uri": {
            "url": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Mac%2FLAST_CHANGE?alt=media",
            "return_content": True,
        },
        "register": "chromium_version_darwin_x86_64",
        "when": "ansible_facts['os_family'] == 'Darwin' and ansible_facts['architecture'] == 'x86_64'",
    },
    {
        "name": "Get latest Chromium version (Darwin arm64)",
        "uri": {
            "url": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Mac_Arm%2FLAST_CHANGE?alt=media",
            "return_content": True,
        },
        "register": "chromium_version_darwin_arm64",
        "when": "ansible_facts['os_family'] == 'Darwin' and ansible_facts['architecture'] == 'arm64'",
    },
    {
        "name": "Download Chromium (Debian)",
        "unarchive": {
            "src": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Linux_x64%2F{{ chromium_version.content }}%2Fchrome-linux.zip?alt=media",
            "remote_src": True,
            "dest": "#{BBOT_TOOLS}",
            "creates": "#{BBOT_TOOLS}/chrome-linux",
        },
        "when": "ansible_facts['os_family'] == 'Debian'",
        "ignore_errors": True,
    },
    {
        "name": "Download Chromium (Darwin x86_64)",
        "unarchive": {
            "src": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Mac%2F{{ chromium_version_darwin_x86_64.content }}%2Fchrome-mac.zip?alt=media",
            "remote_src": True,
            "dest": "#{BBOT_TOOLS}",
            "creates": "#{BBOT_TOOLS}/chrome-mac",
        },
        "when": "ansible_facts['os_family'] == 'Darwin' and ansible_facts['architecture'] == 'x86_64'",
    },
    {
        "name": "Download Chromium (Darwin arm64)",
        "unarchive": {
            "src": "https://www.googleapis.com/download/storage/v1/b/chromium-browser-snapshots/o/Mac_Arm%2F{{ chromium_version_darwin_arm64.content }}%2Fchrome-mac.zip?alt=media",
            "remote_src": True,
            "dest": "#{BBOT_TOOLS}",
            "creates": "#{BBOT_TOOLS}/chrome-mac",
        },
        "when": "ansible_facts['os_family'] == 'Darwin' and ansible_facts['architecture'] == 'arm64'",
    },
    # Because Ubuntu is a special snowflake, we have to bend over backwards to fix the chrome sandbox
    # see https://chromium.googlesource.com/chromium/src/+/main/docs/security/apparmor-userns-restrictions.md
    {
        "name": "Chown chrome_sandbox to root:root",
        "command": {"cmd": "chown -R root:root #{BBOT_TOOLS}/chrome-linux/chrome_sandbox"},
        "when": "ansible_facts['os_family'] == 'Debian'",
        "become": True,
    },
    {
        "name": "Chmod chrome_sandbox to 4755",
        "command": {"cmd": "chmod -R 4755 #{BBOT_TOOLS}/chrome-linux/chrome_sandbox"},
        "when": "ansible_facts['os_family'] == 'Debian'",
        "become": True,
    },
]

DEP_MASSCAN = [
    {
        "name": "install os deps (Debian)",
        "package": {"name": ["gcc", "git", "make", "libpcap0.8-dev"], "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] == 'Debian'",
        "ignore_errors": True,
    },
    {
        "name": "install dev tools (Non-Debian)",
        "package": {"name": ["gcc", "git", "make", "libpcap"], "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] != 'Debian'",
        "ignore_errors": True,
    },
    {
        "name": "Download masscan source code",
        "git": {
            "repo": "https://github.com/robertdavidgraham/masscan.git",
            "dest": "#{BBOT_TEMP}/masscan",
            "single_branch": True,
            "version": "master",
        },
    },
    {
        "name": "Build masscan",
        "command": {
            "chdir": "#{BBOT_TEMP}/masscan",
            "cmd": "make -j",
            "creates": "#{BBOT_TEMP}/masscan/bin/masscan",
        },
    },
    {
        "name": "Install masscan",
        "copy": {"src": "#{BBOT_TEMP}/masscan/bin/masscan", "dest": "#{BBOT_TOOLS}/", "mode": "u+x,g+x,o+x"},
    },
]

DEP_JAVA = [
    {
        "name": "Check if Java is installed",
        "command": "which java",
        "register": "java_installed",
        "ignore_errors": True,
    },
    {
        "name": "Install latest JRE (Debian)",
        "package": {"name": ["default-jre"], "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] == 'Debian' and java_installed.rc != 0",
    },
    {
        "name": "Install latest JRE (Arch)",
        "package": {"name": ["jre-openjdk"], "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] == 'Archlinux' and java_installed.rc != 0",
    },
    {
        "name": "Install latest JRE (Fedora)",
        "package": {"name": ["which", "java-latest-openjdk-headless"], "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] == 'RedHat' and java_installed.rc != 0",
    },
    {
        "name": "Install latest JRE (Alpine)",
        "package": {"name": ["openjdk11"], "state": "present"},
        "become": True,
        "when": "ansible_facts['os_family'] == 'Alpine' and java_installed.rc != 0",
    },
]

# shared module dependencies -- ffuf, massdns, chromium, etc.
SHARED_DEPS = {}
for var, val in list(locals().items()):
    if var.startswith("DEP_") and isinstance(val, list):
        var = var.split("_", 1)[-1].lower()
        SHARED_DEPS[var] = val



import httpx
import logging
from httpx._models import Cookies

log = logging.getLogger("bbot.core.helpers.web.client")


class DummyCookies(Cookies):
    def extract_cookies(self, *args, **kwargs):
        pass


class BBOTAsyncClient(httpx.AsyncClient):
    """
    A subclass of httpx.AsyncClient tailored with BBOT-specific configurations and functionalities.
    This class provides rate limiting, logging, configurable timeouts, user-agent customization, custom
    headers, and proxy settings. Additionally, it allows the disabling of cookies, making it suitable
    for use across an entire scan.

    Attributes:
        _bbot_scan (object): BBOT scan object containing configuration details.
        _persist_cookies (bool): Flag to determine whether cookies should be persisted across requests.

    Examples:
        >>> async with BBOTAsyncClient(_bbot_scan=bbot_scan_object) as client:
        >>>     response = await client.request("GET", "https://example.com")
        >>>     print(response.status_code)
        200
    """

    @classmethod
    def from_config(cls, config, target, *args, **kwargs):
        kwargs["_config"] = config
        kwargs["_target"] = target
        web_config = config.get("web", {})
        retries = kwargs.pop("retries", web_config.get("http_retries", 1))
        ssl_verify = web_config.get("ssl_verify", False)
        if ssl_verify is False:
            from .ssl_context import ssl_context_noverify

            ssl_verify = ssl_context_noverify
        kwargs["transport"] = httpx.AsyncHTTPTransport(retries=retries, verify=ssl_verify)
        kwargs["verify"] = ssl_verify
        return cls(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        self._config = kwargs.pop("_config")
        self._target = kwargs.pop("_target")

        self._web_config = self._config.get("web", {})
        http_debug = self._web_config.get("debug", None)
        if http_debug:
            log.trace(f"Creating AsyncClient: {args}, {kwargs}")

        self._persist_cookies = kwargs.pop("persist_cookies", False)

        # timeout
        http_timeout = self._web_config.get("http_timeout", 20)
        if "timeout" not in kwargs:
            kwargs["timeout"] = http_timeout

        # headers
        headers = kwargs.get("headers", None)
        if headers is None:
            headers = {}

        # cookies
        cookies = kwargs.get("cookies", None)
        if cookies is None:
            cookies = {}

        # user agent
        user_agent = self._web_config.get("user_agent", "BBOT")
        if "User-Agent" not in headers:
            headers["User-Agent"] = user_agent
        kwargs["headers"] = headers
        kwargs["cookies"] = cookies
        # proxy
        proxies = self._web_config.get("http_proxy", None)
        kwargs["proxy"] = proxies

        log.verbose(f"Creating httpx.AsyncClient({args}, {kwargs})")
        super().__init__(*args, **kwargs)
        if not self._persist_cookies:
            self._cookies = DummyCookies()

    def build_request(self, *args, **kwargs):
        if args:
            url = args[0]
            kwargs["url"] = url
        url = kwargs["url"]

        target_in_scope = self._target.in_scope(str(url))

        if target_in_scope:
            if not kwargs.get("cookies", None):
                kwargs["cookies"] = {}
            for ck, cv in self._web_config.get("http_cookies", {}).items():
                if ck not in kwargs["cookies"]:
                    kwargs["cookies"][ck] = cv

        request = super().build_request(**kwargs)

        if target_in_scope:
            for hk, hv in self._web_config.get("http_headers", {}).items():
                hv = str(hv)
                # don't clobber headers
                if hk not in request.headers:
                    request.headers[hk] = hv
        return request

    def _merge_cookies(self, cookies):
        if self._persist_cookies:
            return super()._merge_cookies(cookies)
        return cookies

    @property
    def retries(self):
        return self._transport._pool._retries



import os
import sys
import zmq
import pickle
import asyncio
import inspect
import logging
import tempfile
import traceback
import contextlib
import contextvars
import zmq.asyncio
import multiprocessing
from pathlib import Path
from concurrent.futures import CancelledError
from contextlib import asynccontextmanager, suppress

from bbot.core import CORE
from bbot.errors import BBOTEngineError
from bbot.core.helpers.async_helpers import get_event_loop
from bbot.core.multiprocess import SHARED_INTERPRETER_STATE
from bbot.core.helpers.misc import rand_string, in_exception_chain


error_sentinel = object()


class EngineBase:
    """
    Base Engine class for Server and Client.

    An Engine is a simple and lightweight RPC implementation that allows offloading async tasks
    to a separate process. It leverages ZeroMQ in a ROUTER-DEALER configuration.

    BBOT makes use of this by spawning a dedicated engine for DNS and HTTP tasks.
    This offloads I/O and helps free up the main event loop for other tasks.

    To use Engine, you must subclass both EngineClient and EngineServer.

    See the respective EngineClient and EngineServer classes for usage examples.
    """

    ERROR_CLASS = BBOTEngineError

    def __init__(self, debug=False):
        self._shutdown_status = False
        self.log = logging.getLogger(f"bbot.core.{self.__class__.__name__.lower()}")
        self._engine_debug = debug

    def pickle(self, obj):
        try:
            return pickle.dumps(obj)
        except Exception as e:
            self.log.error(f"Error serializing object: {obj}: {e}")
            self.log.trace(traceback.format_exc())
        return error_sentinel

    def unpickle(self, binary):
        try:
            return pickle.loads(binary)
        except Exception as e:
            self.log.error(f"Error deserializing binary: {e}")
            self.log.trace(f"Offending binary: {binary}")
            self.log.trace(traceback.format_exc())
        return error_sentinel

    async def _infinite_retry(self, callback, *args, **kwargs):
        interval = kwargs.pop("_interval", 300)
        context = kwargs.pop("_context", "")
        # default overall timeout of 10 minutes (300 second interval * 2 iterations)
        max_retries = kwargs.pop("_max_retries", 1)
        if not context:
            context = f"{callback.__name__}({args}, {kwargs})"
        retries = 0
        while not self._shutdown_status:
            try:
                return await asyncio.wait_for(callback(*args, **kwargs), timeout=interval)
            except (TimeoutError, asyncio.exceptions.TimeoutError):
                self.log.debug(f"{self.name}: Timeout after {interval:,} seconds {context}, retrying...")
                retries += 1
                if max_retries is not None and retries > max_retries:
                    raise TimeoutError(f"Timed out after {(max_retries + 1) * interval:,} seconds {context}")

    def engine_debug(self, *args, **kwargs):
        if self._engine_debug:
            self.log.trace(*args, **kwargs)


class EngineClient(EngineBase):
    """
    The client portion of BBOT's RPC Engine.

    To create an engine, you must create a subclass of this class and also
    define methods for each of your desired functions.

    Note that this only supports async functions. If you need to offload a synchronous function to another CPU, use BBOT's multiprocessing pool instead.

    Any CPU or I/O intense logic should be implemented in the EngineServer.

    These functions are typically stubs whose only job is to forward the arguments to the server.

    Functions with the same names should be defined on the EngineServer.

    The EngineClient must specify its associated server class via the `SERVER_CLASS` variable.

    Depending on whether your function is a generator, you will use either `run_and_return()`, or `run_and_yield`.

    Examples:
        >>> from bbot.core.engine import EngineClient
        >>>
        >>> class MyClient(EngineClient):
        >>>     SERVER_CLASS = MyServer
        >>>
        >>>     async def my_function(self, **kwargs)
        >>>         return await self.run_and_return("my_function", **kwargs)
        >>>
        >>>     async def my_generator(self, **kwargs):
        >>>         async for _ in self.run_and_yield("my_generator", **kwargs):
        >>>             yield _
    """

    SERVER_CLASS = None

    def __init__(self, debug=False, **kwargs):
        self.name = f"EngineClient {self.__class__.__name__}"
        super().__init__(debug=debug)
        self.process = None
        if self.SERVER_CLASS is None:
            raise ValueError(f"Must set EngineClient SERVER_CLASS, {self.SERVER_CLASS}")
        self.CMDS = dict(self.SERVER_CLASS.CMDS)
        for k, v in list(self.CMDS.items()):
            self.CMDS[v] = k
        self.socket_address = f"zmq_{rand_string(8)}.sock"
        self.socket_path = Path(tempfile.gettempdir()) / self.socket_address
        self.server_kwargs = kwargs.pop("server_kwargs", {})
        self._server_process = None
        self.context = zmq.asyncio.Context()
        self.context.setsockopt(zmq.LINGER, 0)
        self.sockets = set()

    def check_error(self, message):
        if isinstance(message, dict) and len(message) == 1 and "_e" in message:
            self.engine_debug(f"{self.name}: got error message: {message}")
            error, trace = message["_e"]
            error = self.ERROR_CLASS(error)
            error.engine_traceback = trace
            self.engine_debug(f"{self.name}: raising {error.__class__.__name__}")
            raise error
        return False

    async def run_and_return(self, command, *args, **kwargs):
        fn_str = f"{command}({args}, {kwargs})"
        self.engine_debug(f"{self.name}: executing run-and-return {fn_str}")
        if self._shutdown_status and not command == "_shutdown":
            self.log.verbose(f"{self.name} has been shut down and is not accepting new tasks")
            return
        async with self.new_socket() as socket:
            try:
                message = self.make_message(command, args=args, kwargs=kwargs)
                if message is error_sentinel:
                    return
                await socket.send(message)
                binary = await self._infinite_retry(socket.recv, _context=f"waiting for return value from {fn_str}")
            except BaseException:
                try:
                    await self.send_cancel_message(socket, fn_str)
                except Exception:
                    self.log.debug(f"{self.name}: {fn_str} failed to send cancel message after exception")
                    self.log.trace(traceback.format_exc())
                raise
        # self.log.debug(f"{self.name}.{command}({kwargs}) got binary: {binary}")
        message = self.unpickle(binary)
        self.engine_debug(f"{self.name}: {fn_str} got return value: {message}")
        # error handling
        if self.check_error(message):
            return
        return message

    async def run_and_yield(self, command, *args, **kwargs):
        fn_str = f"{command}({args}, {kwargs})"
        self.engine_debug(f"{self.name}: executing run-and-yield {fn_str}")
        if self._shutdown_status:
            self.log.verbose("Engine has been shut down and is not accepting new tasks")
            return
        message = self.make_message(command, args=args, kwargs=kwargs)
        if message is error_sentinel:
            return
        async with self.new_socket() as socket:
            # TODO: synchronize server-side generator by limiting qsize
            # socket.setsockopt(zmq.RCVHWM, 1)
            # socket.setsockopt(zmq.SNDHWM, 1)
            await socket.send(message)
            while 1:
                try:
                    binary = await self._infinite_retry(
                        socket.recv, _context=f"waiting for new iteration from {fn_str}"
                    )
                    # self.log.debug(f"{self.name}.{command}({kwargs}) got binary: {binary}")
                    message = self.unpickle(binary)
                    self.engine_debug(f"{self.name}: {fn_str} got iteration: {message}")
                    # error handling
                    if self.check_error(message) or self.check_stop(message):
                        break
                    yield message
                except (StopAsyncIteration, GeneratorExit) as e:
                    exc_name = e.__class__.__name__
                    self.engine_debug(f"{self.name}.{command} got {exc_name}")
                    try:
                        await self.send_cancel_message(socket, fn_str)
                    except Exception:
                        self.engine_debug(f"{self.name}.{command} failed to send cancel message after {exc_name}")
                        self.log.trace(traceback.format_exc())
                    break

    async def send_cancel_message(self, socket, context):
        """
        Send a cancel message and wait for confirmation from the server
        """
        # -1 == special "cancel" signal
        message = pickle.dumps({"c": -1})
        await self._infinite_retry(socket.send, message)
        while 1:
            response = await self._infinite_retry(
                socket.recv, _context=f"waiting for CANCEL_OK from {context}", _max_retries=4
            )
            response = pickle.loads(response)
            if isinstance(response, dict):
                response = response.get("m", "")
                if response == "CANCEL_OK":
                    break

    async def send_shutdown_message(self):
        async with self.new_socket() as socket:
            # -99 == special shutdown message
            message = pickle.dumps({"c": -99})
            with suppress(TimeoutError, asyncio.exceptions.TimeoutError):
                await asyncio.wait_for(socket.send(message), 0.5)
            with suppress(TimeoutError, asyncio.exceptions.TimeoutError):
                while 1:
                    response = await asyncio.wait_for(socket.recv(), 0.5)
                    response = pickle.loads(response)
                    if isinstance(response, dict):
                        response = response.get("m", "")
                        if response == "SHUTDOWN_OK":
                            break

    def check_stop(self, message):
        if isinstance(message, dict) and len(message) == 1 and "_s" in message:
            return True
        return False

    def make_message(self, command, args=None, kwargs=None):
        try:
            cmd_id = self.CMDS[command]
        except KeyError:
            raise KeyError(f'Command "{command}" not found. Available commands: {",".join(self.available_commands)}')
        message = {"c": cmd_id}
        if args:
            message["a"] = args
        if kwargs:
            message["k"] = kwargs
        return pickle.dumps(message)

    @property
    def available_commands(self):
        return [s for s in self.CMDS if isinstance(s, str)]

    def start_server(self):
        process_name = multiprocessing.current_process().name
        if SHARED_INTERPRETER_STATE.is_scan_process:
            kwargs = dict(self.server_kwargs)
            # if we're in tests, we use a single event loop to avoid weird race conditions
            # this allows us to more easily mock http, etc.
            if os.environ.get("BBOT_TESTING", "") == "True":
                kwargs["_loop"] = get_event_loop()
            kwargs["debug"] = self._engine_debug
            self.process = CORE.create_process(
                target=self.server_process,
                args=(
                    self.SERVER_CLASS,
                    self.socket_path,
                ),
                kwargs=kwargs,
                custom_name=f"BBOT {self.__class__.__name__}",
            )
            self.process.start()
            return self.process
        else:
            raise BBOTEngineError(
                f"Tried to start server from process {process_name}. Did you forget \"if __name__ == '__main__'?\""
            )

    @staticmethod
    def server_process(server_class, socket_path, **kwargs):
        try:
            loop = kwargs.pop("_loop", None)
            engine_server = server_class(socket_path, **kwargs)
            if loop is not None:
                future = asyncio.run_coroutine_threadsafe(engine_server.worker(), loop)
                future.result()
            else:
                asyncio.run(engine_server.worker())
        except (asyncio.CancelledError, KeyboardInterrupt, CancelledError):
            return
        except Exception:
            import traceback

            log = logging.getLogger("bbot.core.engine.server")
            log.critical(f"Unhandled error in {server_class.__name__} server process: {traceback.format_exc()}")

    @asynccontextmanager
    async def new_socket(self):
        if self._server_process is None:
            self._server_process = self.start_server()
            while not self.socket_path.exists():
                self.engine_debug(f"{self.name}: waiting for server process to start...")
                await asyncio.sleep(0.1)
        socket = self.context.socket(zmq.DEALER)
        socket.setsockopt(zmq.LINGER, 0)  # Discard pending messages immediately disconnect() or close()
        socket.setsockopt(zmq.SNDHWM, 0)  # Unlimited send buffer
        socket.setsockopt(zmq.RCVHWM, 0)  # Unlimited receive buffer
        socket.connect(f"ipc://{self.socket_path}")
        self.sockets.add(socket)
        try:
            yield socket
        finally:
            self.sockets.remove(socket)
            with suppress(Exception):
                socket.close()

    async def shutdown(self):
        if not self._shutdown_status:
            self._shutdown_status = True
            self.log.verbose(f"{self.name}: shutting down...")
            # send shutdown signal
            await self.send_shutdown_message()
            # then terminate context
            try:
                self.context.destroy(linger=0)
            except Exception:
                print(traceback.format_exc(), file=sys.stderr)
            try:
                self.context.term()
            except Exception:
                print(traceback.format_exc(), file=sys.stderr)
            # delete socket file on exit
            self.socket_path.unlink(missing_ok=True)


class EngineServer(EngineBase):
    """
    The server portion of BBOT's RPC Engine.

    Methods defined here must match the methods in your EngineClient.

    To use the functions, you must create mappings for them in the CMDS attribute, as shown below.

    Examples:
        >>> from bbot.core.engine import EngineServer
        >>>
        >>> class MyServer(EngineServer):
        >>>     CMDS = {
        >>>         0: "my_function",
        >>>         1: "my_generator",
        >>>     }
        >>>
        >>>     def my_function(self, arg1=None):
        >>>         await asyncio.sleep(1)
        >>>         return str(arg1)
        >>>
        >>>     def my_generator(self):
        >>>         for i in range(10):
        >>>             await asyncio.sleep(1)
        >>>             yield i
    """

    CMDS = {}

    def __init__(self, socket_path, debug=False):
        self.name = f"EngineServer {self.__class__.__name__}"
        super().__init__(debug=debug)
        self.engine_debug(f"{self.name}: finished setup 1 (_debug={self._engine_debug})")
        self.socket_path = socket_path
        self.client_id_var = contextvars.ContextVar("client_id", default=None)
        # task <--> client id mapping
        self.tasks = {}
        # child tasks spawned by main tasks
        self.child_tasks = {}
        self.engine_debug(f"{self.name}: finished setup 2 (_debug={self._engine_debug})")
        if self.socket_path is not None:
            # create ZeroMQ context
            self.context = zmq.asyncio.Context()
            # ROUTER socket can handle multiple concurrent requests
            self.socket = self.context.socket(zmq.ROUTER)
            self.socket.setsockopt(zmq.LINGER, 0)  # Discard pending messages immediately disconnect() or close()
            self.socket.setsockopt(zmq.SNDHWM, 0)  # Unlimited send buffer
            self.socket.setsockopt(zmq.RCVHWM, 0)  # Unlimited receive buffer
            # create socket file
            self.socket.bind(f"ipc://{self.socket_path}")
        self.engine_debug(f"{self.name}: finished setup 3 (_debug={self._engine_debug})")

    @contextlib.contextmanager
    def client_id_context(self, value):
        token = self.client_id_var.set(value)
        try:
            yield
        finally:
            self.client_id_var.reset(token)

    async def run_and_return(self, client_id, command_fn, *args, **kwargs):
        fn_str = f"{command_fn.__name__}({args}, {kwargs})"
        self.engine_debug(fn_str)
        with self.client_id_context(client_id):
            try:
                self.engine_debug(f"{self.name}: starting run-and-return {fn_str}")
                try:
                    result = await command_fn(*args, **kwargs)
                except BaseException as e:
                    if in_exception_chain(e, (KeyboardInterrupt, asyncio.CancelledError)):
                        log_fn = self.log.debug
                    else:
                        log_fn = self.log.error
                    error = f"{self.name}: error in {fn_str}: {e}"
                    trace = traceback.format_exc()
                    log_fn(error)
                    self.log.trace(trace)
                    result = {"_e": (error, trace)}
                finally:
                    self.tasks.pop(client_id, None)
                    self.engine_debug(f"{self.name}: sending response to {fn_str}: {result}")
                    await self.send_socket_multipart(client_id, result)
            except BaseException as e:
                self.log.critical(
                    f"Unhandled exception in {self.name}.run_and_return({client_id}, {command_fn}, {args}, {kwargs}): {e}"
                )
                self.log.critical(traceback.format_exc())
            finally:
                self.engine_debug(f"{self.name} finished run-and-return {fn_str}")

    async def run_and_yield(self, client_id, command_fn, *args, **kwargs):
        fn_str = f"{command_fn.__name__}({args}, {kwargs})"
        with self.client_id_context(client_id):
            try:
                self.engine_debug(f"{self.name}: starting run-and-yield {fn_str}")
                try:
                    async for _ in command_fn(*args, **kwargs):
                        self.engine_debug(f"{self.name}: sending iteration for {fn_str}: {_}")
                        await self.send_socket_multipart(client_id, _)
                except BaseException as e:
                    if in_exception_chain(e, (KeyboardInterrupt, asyncio.CancelledError)):
                        log_fn = self.log.debug
                    else:
                        log_fn = self.log.error
                    error = f"{self.name}: error in {fn_str}: {e}"
                    trace = traceback.format_exc()
                    log_fn(error)
                    self.log.trace(trace)
                    result = {"_e": (error, trace)}
                    await self.send_socket_multipart(client_id, result)
                finally:
                    self.engine_debug(f"{self.name}: reached end of run-and-yield iteration for {fn_str}")
                    # _s == special signal that means StopIteration
                    await self.send_socket_multipart(client_id, {"_s": None})
                    self.tasks.pop(client_id, None)
            except BaseException as e:
                self.log.critical(
                    f"Unhandled exception in {self.name}.run_and_yield({client_id}, {command_fn}, {args}, {kwargs}): {e}"
                )
                self.log.critical(traceback.format_exc())
            finally:
                self.engine_debug(f"{self.name}: finished run-and-yield {fn_str}")

    async def send_socket_multipart(self, client_id, message):
        try:
            message = pickle.dumps(message)
            await self._infinite_retry(self.socket.send_multipart, [client_id, message])
        except Exception as e:
            self.log.verbose(f"{self.name}: error sending ZMQ message: {e}")
            self.log.trace(traceback.format_exc())

    def check_error(self, message):
        if message is error_sentinel:
            return True

    async def worker(self):
        self.engine_debug(f"{self.name}: starting worker")
        try:
            while 1:
                client_id, binary = await self.socket.recv_multipart()
                message = self.unpickle(binary)
                self.engine_debug(f"{self.name} got message: {message}")
                if self.check_error(message):
                    continue

                cmd = message.get("c", None)
                if not isinstance(cmd, int):
                    self.log.warning(f"{self.name}: no command sent in message: {message}")
                    continue

                # -1 == cancel task
                if cmd == -1:
                    self.engine_debug(f"{self.name} got cancel signal")
                    await self.send_socket_multipart(client_id, {"m": "CANCEL_OK"})
                    await self.cancel_task(client_id)
                    continue

                # -99 == shutdown task
                if cmd == -99:
                    self.log.verbose(f"{self.name} got shutdown signal")
                    await self.send_socket_multipart(client_id, {"m": "SHUTDOWN_OK"})
                    await self._shutdown()
                    return

                args = message.get("a", ())
                if not isinstance(args, tuple):
                    self.log.warning(f"{self.name}: received invalid args of type {type(args)}, should be tuple")
                    continue
                kwargs = message.get("k", {})
                if not isinstance(kwargs, dict):
                    self.log.warning(f"{self.name}: received invalid kwargs of type {type(kwargs)}, should be dict")
                    continue

                command_name = self.CMDS[cmd]
                command_fn = getattr(self, command_name, None)

                if command_fn is None:
                    self.log.warning(f'{self.name} has no function named "{command_fn}"')
                    continue

                if inspect.isasyncgenfunction(command_fn):
                    self.engine_debug(f"{self.name}: creating run-and-yield coroutine for {command_name}()")
                    coroutine = self.run_and_yield(client_id, command_fn, *args, **kwargs)
                else:
                    self.engine_debug(f"{self.name}: creating run-and-return coroutine for {command_name}()")
                    coroutine = self.run_and_return(client_id, command_fn, *args, **kwargs)

                self.engine_debug(f"{self.name}: creating task for {command_name}() coroutine")
                task = asyncio.create_task(coroutine)
                self.tasks[client_id] = task, command_fn, args, kwargs
                self.engine_debug(f"{self.name}: finished creating task for {command_name}() coroutine")
        except BaseException as e:
            await self._shutdown()
            if not in_exception_chain(e, (KeyboardInterrupt, asyncio.CancelledError)):
                self.log.error(f"{self.name}: error in EngineServer worker: {e}")
                self.log.trace(traceback.format_exc())
        finally:
            self.engine_debug(f"{self.name}: finished worker()")

    async def _shutdown(self):
        if not self._shutdown_status:
            self.log.verbose(f"{self.name}: shutting down...")
            self._shutdown_status = True
            await self.cancel_all_tasks()
            context = getattr(self, "context", None)
            if context is not None:
                try:
                    context.destroy(linger=0)
                except Exception:
                    self.log.trace(traceback.format_exc())
                try:
                    context.term()
                except Exception:
                    self.log.trace(traceback.format_exc())
            self.log.verbose(f"{self.name}: finished shutting down")

    async def task_pool(self, fn, args_kwargs, threads=10, timeout=300, global_kwargs=None):
        if global_kwargs is None:
            global_kwargs = {}

        tasks = {}
        args_kwargs = list(args_kwargs)

        def new_task():
            if args_kwargs:
                kwargs = {}
                tracker = None
                args = args_kwargs.pop(0)
                if isinstance(args, (list, tuple)):
                    # you can specify a custom tracker value if you want
                    # this helps with correlating results
                    with suppress(ValueError):
                        args, kwargs, tracker = args
                    # or you can just specify args/kwargs
                    with suppress(ValueError):
                        args, kwargs = args

                if not isinstance(kwargs, dict):
                    raise ValueError(f"kwargs must be dict (got: {kwargs})")
                if not isinstance(args, (list, tuple)):
                    args = [args]

                task = self.new_child_task(fn(*args, **kwargs, **global_kwargs))
                tasks[task] = (args, kwargs, tracker)

        for _ in range(threads):  # Start initial batch of tasks
            new_task()

        while tasks:  # While there are tasks pending
            # Wait for the first task to complete
            finished = await self.finished_tasks(tasks, timeout=timeout)
            for task in finished:
                result = task.result()
                (args, kwargs, tracker) = tasks.pop(task)
                yield (args, kwargs, tracker), result
                new_task()

    def new_child_task(self, coro):
        """
        Create a new asyncio task, making sure to track it based on the client id.

        This allows the task to be automatically cancelled if its parent is cancelled.
        """
        client_id = self.client_id_var.get()
        task = asyncio.create_task(coro)

        if client_id:

            def remove_task(t):
                tasks = self.child_tasks.get(client_id, set())
                tasks.discard(t)
                if not tasks:
                    self.child_tasks.pop(client_id, None)

            task.add_done_callback(remove_task)

            try:
                self.child_tasks[client_id].add(task)
            except KeyError:
                self.child_tasks[client_id] = {task}

        return task

    async def finished_tasks(self, tasks, timeout=None):
        """
        Given a list of asyncio tasks, return the ones that are finished with an optional timeout
        """
        if tasks:
            try:
                done, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED, timeout=timeout)
                return done
            except BaseException as e:
                if isinstance(e, (TimeoutError, asyncio.exceptions.TimeoutError)):
                    self.log.warning(f"{self.name}: Timeout after {timeout:,} seconds in finished_tasks({tasks})")
                    for task in list(tasks):
                        task.cancel()
                        self._await_cancelled_task(task)
                else:
                    if not in_exception_chain(e, (KeyboardInterrupt, asyncio.CancelledError)):
                        self.log.error(f"{self.name}: Unhandled exception in finished_tasks({tasks}): {e}")
                        self.log.trace(traceback.format_exc())
                    raise
        return set()

    async def cancel_task(self, client_id):
        parent_task = self.tasks.pop(client_id, None)
        if parent_task is None:
            return
        parent_task, _cmd, _args, _kwargs = parent_task
        self.engine_debug(f"{self.name}: Cancelling client id {client_id} (task: {parent_task})")
        parent_task.cancel()
        child_tasks = self.child_tasks.pop(client_id, set())
        if child_tasks:
            self.engine_debug(f"{self.name}: Cancelling {len(child_tasks):,} child tasks for client id {client_id}")
            for child_task in child_tasks:
                child_task.cancel()

        for task in [parent_task] + list(child_tasks):
            await self._await_cancelled_task(task)

    async def _await_cancelled_task(self, task):
        try:
            await asyncio.wait_for(task, timeout=10)
        except (TimeoutError, asyncio.exceptions.TimeoutError):
            self.log.trace(f"{self.name}: Timeout cancelling task: {task}")
            return
        except (KeyboardInterrupt, asyncio.CancelledError):
            return
        except BaseException as e:
            self.log.error(f"Unhandled error in {task.get_coro().__name__}(): {e}")
            self.log.trace(traceback.format_exc())

    async def cancel_all_tasks(self):
        for client_id in list(self.tasks):
            await self.cancel_task(client_id)
        for client_id, tasks in self.child_tasks.items():
            for task in list(tasks):
                await self._await_cancelled_task(task)



from .base import ModuleTestBase


class TestOTX(ModuleTestBase):
    config_overrides = {"modules": {"otx": {"api_key": "test"}}}

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://otx.alienvault.com/api/v1/indicators/domain/blacklanternsecurity.com/passive_dns",
            json={
                "passive_dns": [
                    {
                        "address": "2606:50c0:8000::153",
                        "first": "2021-10-28T20:23:08",
                        "last": "2022-08-24T18:29:49",
                        "hostname": "asdf.blacklanternsecurity.com",
                        "record_type": "AAAA",
                        "indicator_link": "/indicator/hostname/www.blacklanternsecurity.com",
                        "flag_url": "assets/images/flags/us.png",
                        "flag_title": "United States",
                        "asset_type": "hostname",
                        "asn": "AS54113 fastly",
                    }
                ]
            },
            headers={"X-OTX-API-KEY": "test"},
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from .base import ModuleTestBase


class TestUrl_Manipulation(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "url_manipulation"]
    body = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello null!</p>';
    </body>
    </html>
    """

    body_match = """
    <html>
    <title>the title</title>
    <body>
    <p>Hello AAAAAAAAAAAAAA!</p>';
    </body>
    </html>
    """

    async def setup_after_prep(self, module_test):
        expect_args = {"query_string": f"{module_test.module.rand_string}=.xml".encode()}
        respond_args = {"response_data": self.body_match}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        respond_args = {"response_data": self.body}
        module_test.set_expect_requests(respond_args=respond_args)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == f"Url Manipulation: [body] Sig: [Modified URL: http://127.0.0.1:8888/?{module_test.module.rand_string}=.xml]"
            for e in events
        )



from .base import ModuleTestBase

from bbot.scanner import Scanner


class TestCloudCheck(ModuleTestBase):
    targets = ["http://127.0.0.1:8888", "asdf2.storage.googleapis.com"]
    modules_overrides = ["httpx", "excavate", "cloudcheck"]

    async def setup_after_prep(self, module_test):
        module_test.set_expect_requests({"uri": "/"}, {"response_data": "<a href='http://asdf.s3.amazonaws.com'/>"})

        scan = Scanner(config={"cloudcheck": True})
        await scan._prep()
        module = scan.modules["cloudcheck"]
        from cloudcheck import providers

        # make sure we have at least one provider
        assert providers.Amazon.name == "Amazon"

        ip_event = scan.make_event("8.8.8.8", parent=scan.root_event)
        aws_event1 = scan.make_event("amazonaws.com", parent=scan.root_event)
        aws_event2 = scan.make_event("asdf.amazonaws.com", parent=scan.root_event)
        aws_event3 = scan.make_event("asdfamazonaws.com", parent=scan.root_event)
        aws_event4 = scan.make_event("test.asdf.aws", parent=scan.root_event)

        other_event1 = scan.make_event("cname.evilcorp.com", parent=scan.root_event)
        other_event2 = scan.make_event("cname2.evilcorp.com", parent=scan.root_event)
        other_event3 = scan.make_event("cname3.evilcorp.com", parent=scan.root_event)
        other_event2._resolved_hosts = {"8.8.8.8"}
        other_event3._resolved_hosts = {"asdf.amazonaws.com"}

        for event in (ip_event, other_event2):
            await module.handle_event(ip_event)
            assert "cloud-google" in ip_event.tags
            assert "google-ip" in ip_event.tags

        for event in (aws_event1, aws_event2, aws_event4, other_event3):
            await module.handle_event(event)
            assert "cloud-amazon" in event.tags, f"{event} was not properly cloud-tagged"

        assert "amazon-domain" in aws_event1.tags
        assert "amazon-cname" in other_event3.tags

        for event in (aws_event3, other_event1):
            await module.handle_event(event)
            assert "cloud-amazon" not in event.tags, f"{event} was improperly cloud-tagged"
            assert not any(t for t in event.tags if t.startswith("cloud-") or t.startswith("cdn-")), (
                f"{event} was improperly cloud-tagged"
            )

        google_event1 = scan.make_event("asdf.googleapis.com", parent=scan.root_event)
        google_event2 = scan.make_event("asdf.google", parent=scan.root_event)
        google_event3 = scan.make_event("asdf.evilcorp.com", parent=scan.root_event)
        google_event3._resolved_hosts = {"asdf.storage.googleapis.com"}

        for event in (google_event1, google_event2, google_event3):
            await module.handle_event(event)
            assert "cloud-google" in event.tags, f"{event} was not properly cloud-tagged"

        await scan._cleanup()

    def check(self, module_test, events):
        assert 2 == len([e for e in events if e.type == "STORAGE_BUCKET"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "STORAGE_BUCKET"
                and e.data["name"] == "asdf"
                and str(e.module) == "cloudcheck"
                and "cloud-amazon" in e.tags
                and "amazon-domain" in e.tags
                and e.scope_distance == 1
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "STORAGE_BUCKET"
                and e.data["name"] == "asdf2"
                and str(e.module) == "cloudcheck"
                and "cloud-google" in e.tags
                and "google-domain" in e.tags
                and e.scope_distance == 0
            ]
        )



from bbot.modules.base import BaseModule


class azure_tenant(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["affiliates", "subdomain-enum", "cloud-enum", "passive", "safe"]
    meta = {
        "description": "Query Azure via azmap.dev for tenant sister domains",
        "created_date": "2024-07-04",
        "author": "@TheTechromancer",
    }

    base_url = "https://azmap.dev/api/tenant"
    in_scope_only = True
    per_domain_only = True

    async def setup(self):
        self.processed = set()
        return True

    async def handle_event(self, event):
        _, query = self.helpers.split_domain(event.data)
        tenant_data = await self.query(query)

        if not tenant_data:
            return

        tenant_id = tenant_data.get("tenant_id")
        tenant_name = tenant_data.get("tenant_name")
        email_domains = tenant_data.get("email_domains", [])

        if email_domains:
            self.verbose(
                f'Found {len(email_domains):,} domains under tenant for "{query}": {", ".join(sorted(email_domains))}'
            )
            for domain in email_domains:
                if domain != query:
                    await self.emit_event(
                        domain,
                        "DNS_NAME",
                        parent=event,
                        tags=["affiliate", "azure-tenant"],
                        context=f'{{module}} queried azmap.dev for "{query}" and found {{event.type}}: {{event.data}}',
                    )

            # Build tenant names list (include the tenant name from the API)
            tenant_names = []
            if tenant_name:
                tenant_names.append(tenant_name)

            # Also extract tenant names from .onmicrosoft.com domains
            for domain in email_domains:
                if domain.lower().endswith(".onmicrosoft.com"):
                    tenantname = domain.split(".")[0].lower()
                    if tenantname and tenantname not in tenant_names:
                        tenant_names.append(tenantname)

            event_data = {"tenant-names": tenant_names, "domains": sorted(email_domains)}
            tenant_names_str = ",".join(tenant_names)
            if tenant_id:
                event_data["tenant-id"] = tenant_id
            await self.emit_event(
                event_data,
                "AZURE_TENANT",
                parent=event,
                context=f'{{module}} queried azmap.dev for "{query}" and found {{event.type}}: {tenant_names_str}',
            )

    async def query(self, domain):
        url = f"{self.base_url}?domain={domain}&extract=true"

        self.debug(f"Retrieving tenant domains at {url}")

        r = await self.helpers.request(url)
        status_code = getattr(r, "status_code", 0)
        if status_code != 200:
            self.verbose(f'Error retrieving azure_tenant domains for "{domain}" (status code: {status_code})')
            return {}

        try:
            tenant_data = r.json()
        except Exception as e:
            self.warning(f'Error parsing JSON response for "{domain}": {e}')
            return {}

        # Absorb domains into word cloud
        email_domains = tenant_data.get("email_domains", [])
        for d in email_domains:
            d = str(d).lower()
            _, query = self.helpers.split_domain(d)
            self.processed.add(hash(query))
            self.scan.word_cloud.absorb_word(d)

        return tenant_data



import pytest
import random
import string
from bbot.core.helpers.misc import make_ip_type, is_ip


class TestIPAddressBenchmarks:
    """
    Benchmark tests for IP address processing operations.

    These tests measure the performance of BBOT-level IP functions which are
    critical for network scanning efficiency and could benefit from different
    underlying implementations.
    """

    def setup_method(self):
        """Setup common test data"""
        # Set deterministic seed for consistent benchmark results
        random.seed(42)  # Fixed seed for reproducible results

        # Generate test data of different types and sizes
        self.valid_ips = self._generate_valid_ips()
        self.invalid_ips = self._generate_invalid_ips()
        self.mixed_data = self._generate_mixed_data()

    def _generate_valid_ips(self):
        """Generate valid IP addresses for testing"""
        valid_ips = []

        # IPv4 addresses
        for i in range(1000):
            valid_ips.append(
                f"{random.randint(1, 223)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(1, 254)}"
            )

        # IPv6 addresses
        for i in range(500):
            ipv6_parts = []
            for j in range(8):
                ipv6_parts.append(f"{random.randint(0, 65535):x}")
            valid_ips.append(":".join(ipv6_parts))

        # Network addresses
        for i in range(500):
            base_ip = f"{random.randint(1, 223)}.{random.randint(0, 255)}.{random.randint(0, 255)}.0"
            valid_ips.append(f"{base_ip}/{random.randint(8, 30)}")

        # IP ranges
        for i in range(200):
            start_ip = (
                f"{random.randint(1, 223)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(1, 200)}"
            )
            end_ip = f"{random.randint(1, 223)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(201, 254)}"
            valid_ips.append(f"{start_ip}-{end_ip}")

        return valid_ips

    def _generate_invalid_ips(self):
        """Generate invalid IP addresses for testing"""
        invalid_ips = []

        # Malformed IPv4
        for i in range(500):
            invalid_ips.append(
                f"{random.randint(256, 999)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}"
            )
            invalid_ips.append(f"{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}")
            invalid_ips.append(
                f"{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}.{random.randint(0, 255)}"
            )

        # Malformed IPv6
        for i in range(300):
            ipv6_parts = []
            for j in range(random.randint(5, 10)):  # Wrong number of parts
                ipv6_parts.append(f"{random.randint(0, 65535):x}")
            invalid_ips.append(":".join(ipv6_parts))

        # Random strings
        for i in range(200):
            length = random.randint(5, 20)
            invalid_ips.append("".join(random.choices(string.ascii_letters + string.digits, k=length)))

        return invalid_ips

    def _generate_mixed_data(self):
        """Generate mixed valid/invalid data for realistic testing"""
        mixed = []
        mixed.extend(self.valid_ips[:500])  # First 500 valid
        mixed.extend(self.invalid_ips[:500])  # First 500 invalid
        # Use deterministic shuffle with fixed seed for consistent results
        random.seed(42)  # Reset seed before shuffle
        random.shuffle(mixed)  # Shuffle for realistic distribution
        return mixed

    @pytest.mark.benchmark(group="ip_validation")
    def test_is_ip_performance(self, benchmark):
        """Benchmark IP validation performance with mixed data"""

        def validate_ips():
            valid_count = 0
            for ip in self.mixed_data:
                if is_ip(ip):
                    valid_count += 1
            return valid_count

        result = benchmark(validate_ips)
        assert result > 0

    @pytest.mark.benchmark(group="ip_type_detection")
    def test_make_ip_type_performance(self, benchmark):
        """Benchmark IP type detection performance"""

        def detect_ip_types():
            type_count = 0
            for ip in self.valid_ips:
                try:
                    make_ip_type(ip)
                    type_count += 1
                except Exception:
                    pass
            return type_count

        result = benchmark(detect_ip_types)
        assert result > 0

    @pytest.mark.benchmark(group="ip_processing")
    def test_mixed_ip_operations(self, benchmark):
        """Benchmark combined IP validation + type detection"""

        def process_ips():
            processed = 0
            for ip in self.mixed_data:
                if is_ip(ip):
                    try:
                        make_ip_type(ip)
                        processed += 1
                    except Exception:
                        pass
            return processed

        result = benchmark(process_ips)
        assert result > 0



from pathlib import Path
from .base import ModuleTestBase, tempwordlist
import pytest


@pytest.fixture(params=["ssh", "ftp", "telnet", "vnc", "mssql", "mysql", "postgresql"])
def protocol(request):
    return request.param


@pytest.fixture
def mock_legba_run_process(monkeypatch, request):
    async def fake_run_process(self, cmd):
        try:
            # find index of `--output` in cmd
            output_index = cmd.index("--output")
            # output_path is directly after `--output` in cmd
            output_path = Path(cmd[output_index + 1])
        except Exception as e:
            raise Exception(f"Could not determine output file path from command {cmd}: {e}")

        protocol = request.getfixturevalue("protocol")

        expected_file_content_per_protocol = {
            "ssh": '{"found_at":"2025-07-22T20:50:19.541305293+02:00","target":"127.0.0.1:2222","plugin":"ssh","data":{"username":"remnux","password":"malware"},"partial":false}',
            "ftp": '{"found_at":"2025-07-22T20:51:19.541305293+02:00","target":"127.0.0.1:21","plugin":"ftp","data":{"username":"ftp_boot","password":"ftp_boot"},"partial":false}',
            "telnet": '{"found_at":"2025-07-22T20:51:19.541305293+02:00","target":"127.0.0.1:23","plugin":"telnet","data":{"username":"guest","password":"guest"},"partial":false}',
            "vnc": '{"found_at":"2025-07-22T20:51:19.541305293+02:00","target":"127.0.0.1:5900","plugin":"vnc","data":{"username":"Administrator","password":""},"partial":false}',
            "mssql": '{"found_at":"2025-07-22T20:51:19.541305293+02:00","target":"127.0.0.1:1433","plugin":"mssql","data":{"username":"sa","password":"default"},"partial":false}',
            "mysql": '{"found_at":"2025-07-22T20:51:19.541305293+02:00","target":"127.0.0.1:3306","plugin":"mysql","data":{"username":"root","password":"moves"},"partial":false}',
            "postgresql": '{"found_at":"2025-07-22T20:51:19.541305293+02:00","target":"127.0.0.1:5432","plugin":"pgsql","data":{"username":"postgres","password":"postgres"},"partial":false}',
        }

        output_path.write_text(expected_file_content_per_protocol[protocol])

    from bbot.modules.base import BaseModule

    monkeypatch.setattr(BaseModule, "run_process", fake_run_process)


@pytest.mark.usefixtures("mock_legba_run_process")
class TestLegba(ModuleTestBase):
    targets = ["127.0.0.1"]

    temp_ssh_wordlist = tempwordlist(["test:test", "admin:admin", "admin:password", "remnux:malware", "user:pass"])
    temp_ftp_wordlist = tempwordlist(["test:test", "ftp_boot:ftp_boot", "admin:password", "root:root", "user:pass"])
    temp_telnet_wordlist = tempwordlist(["test:test", "admin:admin", "admin:password", "root:root", "guest:guest"])
    temp_vnc_wordlist = tempwordlist(["test", "admin", "password", "Administrator", "pass"])
    temp_mssql_wordlist = tempwordlist(["sa:default", "admin:admin", "admin:password", "root:root", "user:pass"])
    temp_mysql_wordlist = tempwordlist(["test:test", "admin:admin", "root:moves", "root:root", "user:pass"])
    temp_postgresql_wordlist = tempwordlist(["postgres:postgres", "admin:admin", "admin:password", "user:pass"])

    config_overrides = {
        "modules": {
            "legba": {
                "ssh_wordlist": str(temp_ssh_wordlist),
                "ftp_wordlist": str(temp_ftp_wordlist),
                "telnet_wordlist": str(temp_telnet_wordlist),
                "vnc_wordlist": str(temp_vnc_wordlist),
                "mssql_wordlist": str(temp_mssql_wordlist),
                "mysql_wordlist": str(temp_mysql_wordlist),
                "postgresql_wordlist": str(temp_postgresql_wordlist),
            }
        }
    }

    @pytest.fixture(autouse=True)
    def _protocol_dependency(self, protocol):
        # ensure pytest sees dependency and runs one test per protocol
        self._protocol = protocol

    async def setup_after_prep(self, module_test):
        protocol = module_test.request_fixture.getfixturevalue("protocol")
        ports = {"ssh": 2222, "ftp": 21, "telnet": 23, "vnc": 5900, "mssql": 1433, "mysql": 3306, "postgresql": 5432}
        event_data = {"host": str(self.targets[0]), "protocol": protocol.upper(), "port": ports[protocol]}
        protocol_event = module_test.scan.make_event(
            event_data,
            "PROTOCOL",
            parent=module_test.scan.root_event,
        )

        await module_test.module.emit_event(protocol_event)

    def check(self, module_test, events):
        protocol = module_test.request_fixture.getfixturevalue("protocol")
        finding_events = [e for e in events if e.type == "FINDING"]

        assert len(finding_events) == 1

        expected_desc = {
            "ssh": "Valid ssh credentials found - remnux:malware",
            "ftp": "Valid ftp credentials found - ftp_boot:ftp_boot",
            "telnet": "Valid telnet credentials found - guest:guest",
            "vnc": "Valid vnc credentials found - Administrator",
            "mssql": "Valid mssql credentials found - sa:default",
            "mysql": "Valid mysql credentials found - root:moves",
            "postgresql": "Valid postgresql credentials found - postgres:postgres",
        }

        assert expected_desc[protocol] in finding_events[0].data["description"]



import json

from bbot.modules.templates.subdomain_enum import subdomain_enum


class dnsdumpster(subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query dnsdumpster for subdomains",
        "created_date": "2022-03-12",
        "author": "@TheTechromancer",
    }

    base_url = "https://dnsdumpster.com"

    async def setup(self):
        self.apikey_regex = self.helpers.re.compile(r'<form[^>]*data-form-id="mainform"[^>]*hx-headers=\'([^\']*)\'')
        return True

    async def query(self, domain):
        ret = []
        # first, get the JWT token from the main page
        res1 = await self.api_request(self.base_url)
        status_code = getattr(res1, "status_code", 0)
        if status_code not in [200]:
            self.verbose(f'Bad response code "{status_code}" from DNSDumpster')
            return ret

        # Extract JWT token from the form's hx-headers attribute using regex
        jwt_token = None
        try:
            # Look for the form with data-form-id="mainform" and extract hx-headers
            form_match = await self.helpers.re.search(self.apikey_regex, res1.text)
            if form_match:
                headers_json = form_match.group(1)
                headers_data = json.loads(headers_json)
                jwt_token = headers_data.get("Authorization")
        except (AttributeError, json.JSONDecodeError, KeyError):
            self.log.warning("Error obtaining JWT token")
            return ret

        # Abort if we didn't get the JWT token
        if not jwt_token:
            self.verbose("Error obtaining JWT token")
            self.errorState = True
            return ret
        else:
            self.debug("Successfully obtained JWT token")

        if self.scan.stopping:
            return ret

        # Query the API with the JWT token
        res2 = await self.api_request(
            "https://api.dnsdumpster.com/htmld/",
            method="POST",
            data={"target": str(domain).lower()},
            headers={
                "Authorization": jwt_token,
                "Content-Type": "application/x-www-form-urlencoded",
                "Origin": "https://dnsdumpster.com",
                "Referer": "https://dnsdumpster.com/",
                "HX-Request": "true",
                "HX-Target": "results",
                "HX-Current-URL": "https://dnsdumpster.com/",
            },
        )
        status_code = getattr(res2, "status_code", 0)
        if status_code not in [200]:
            self.verbose(f'Bad response code "{status_code}" from DNSDumpster API')
            return ret

        return await self.scan.extract_in_scope_hostnames(res2.text)



from bbot.modules.templates.bucket import bucket_template


class bucket_microsoft(bucket_template):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic"]
    meta = {
        "description": "Check for Azure storage blobs related to target",
        "created_date": "2022-11-04",
        "author": "@TheTechromancer",
    }
    options = {"permutations": False}
    options_desc = {
        "permutations": "Whether to try permutations",
    }

    cloudcheck_provider_name = "Microsoft"
    delimiters = ("", "-")
    base_domains = ["blob.core.windows.net"]
    # Dirbusting is required to know whether a bucket is public
    supports_open_check = False

    def build_bucket_request(self, bucket_name, base_domain, region):
        url = self.build_url(bucket_name, base_domain, region)
        url = url.strip("/") + f"/{bucket_name}?restype=container"
        return url, {}

    def check_bucket_exists(self, bucket_name, response):
        status_code = getattr(response, "status_code", 0)
        existent_bucket = status_code != 0
        return existent_bucket, set()

    def clean_bucket_url(self, url):
        # only return root URL
        return "/".join(url.split("/")[:3])



import asyncio
import regex as re
from contextlib import suppress

from bbot.modules.base import BaseInterceptModule


class CloudCheck(BaseInterceptModule):
    watched_events = ["*"]
    meta = {
        "description": "Tag events by cloud provider, identify cloud resources like storage buckets",
        "created_date": "2024-07-07",
        "author": "@TheTechromancer",
    }
    # tag events up to and including distance-2
    scope_distance_modifier = 2
    _priority = 3

    async def setup(self):
        self._cloud_hostname_regexes = None
        self._cloud_hostname_regexes_lock = asyncio.Lock()
        # perform a test lookup during setup to force signature update
        await self.helpers.cloudcheck.lookup("8.8.8.8")
        return True

    async def filter_event(self, event):
        if (not event.host) or (event.type in ("IP_RANGE",)):
            return False, "event does not have host attribute"
        return True

    async def handle_event(self, event, **kwargs):
        # cloud tagging by hosts
        hosts_to_check = set(event.resolved_hosts)
        with suppress(KeyError):
            hosts_to_check.remove(event.host_original)
        hosts_to_check = [str(event.host_original)] + list(hosts_to_check)

        for i, host in enumerate(hosts_to_check):
            host_is_ip = self.helpers.is_ip(host)
            try:
                cloudcheck_results = await self.helpers.cloudcheck.lookup(host)
            except Exception as e:
                self.warning(f"Error running cloudcheck against {event} (host: {host}): {e}")
                continue
            for provider in cloudcheck_results:
                provider_name = provider["name"].lower()
                tags = provider.get("tags", [])
                for tag in tags:
                    event.add_tag(tag)
                    event.add_tag(f"{tag}-{provider_name}")
                    if host_is_ip:
                        event.add_tag(f"{provider_name}-ip")
                    else:
                        # if the original hostname is a cloud domain, tag it as such
                        if i == 0:
                            event.add_tag(f"{provider_name}-domain")
                        # any children are tagged as CNAMEs
                        else:
                            event.add_tag(f"{provider_name}-cname")

        # we only generate storage buckets off of in-scope or distance-1 events
        if event.scope_distance >= self.max_scope_distance:
            return

        # see if any of our hosts are storage buckets, etc.
        regexes = await self.cloud_hostname_regexes()
        regexes = regexes.get("STORAGE_BUCKET_HOSTNAME", [])
        for regex_name, regex in regexes.items():
            for host in hosts_to_check:
                if match := regex.match(host):
                    try:
                        bucket_name, bucket_domain = match.groups()
                    except Exception as e:
                        self.error(
                            f"Bucket regex {regex_name} ({regex}) is not formatted correctly to extract bucket name and domain: {e}"
                        )
                        continue
                    bucket_name, bucket_domain = match.groups()
                    bucket_url = f"https://{bucket_name}.{bucket_domain}"
                    await self.emit_event(
                        {
                            "name": bucket_name,
                            "url": bucket_url,
                            "context": f"{{module}} analyzed {event.type} and found {{event.type}}: {bucket_url}",
                        },
                        "STORAGE_BUCKET",
                        parent=event,
                    )

    async def cloud_hostname_regexes(self):
        async with self._cloud_hostname_regexes_lock:
            if not self._cloud_hostname_regexes:
                storage_bucket_regexes = {}
                self._cloud_hostname_regexes = {"STORAGE_BUCKET_HOSTNAME": storage_bucket_regexes}
                from cloudcheck import providers

                for attr in dir(providers):
                    if attr.startswith("_"):
                        continue
                    provider = getattr(providers, attr)
                    provider_regexes = getattr(provider, "regexes", {})
                    for regex_name, regexes in provider_regexes.items():
                        for i, regex in enumerate(regexes):
                            if not regex_name in ("STORAGE_BUCKET_HOSTNAME"):
                                continue
                            try:
                                storage_bucket_regexes[f"{attr}-{regex_name}-{i}"] = re.compile(regex)
                            except Exception as e:
                                self.error(f"Error compiling regex for {attr}-{regex_name}: {e}")
                                continue
            return self._cloud_hostname_regexes



from pathlib import Path


def sanitize_git_repo(repo_folder: Path):
    # sanitizing the git config is infeasible since there are too many different ways to do evil things
    # instead, we move it out of .git and into the repo folder, so we don't miss any secrets etc. inside
    config_file = repo_folder / ".git" / "config"
    if config_file.exists():
        config_file.rename(repo_folder / "git_config_original")
    # move the index file
    index_file = repo_folder / ".git" / "index"
    if index_file.exists():
        index_file.rename(repo_folder / "git_index_original")
    # move the hooks folder
    hooks_folder = repo_folder / ".git" / "hooks"
    if hooks_folder.exists():
        hooks_folder.rename(repo_folder / "git_hooks_original")



import logging

from bbot.errors import *

log = logging.getLogger("bbot.preset.conditions")

JINJA_ENV = None


class ConditionEvaluator:
    def __init__(self, preset):
        self.preset = preset

    @property
    def context(self):
        return {
            "preset": self.preset,
            "config": self.preset.config,
            "abort": self.abort,
            "warn": self.warn,
        }

    def abort(self, message):
        if not self.preset.force_start:
            raise PresetAbortError(message)

    def warn(self, message):
        log.warning(message)

    def evaluate(self):
        context = self.context
        already_evaluated = set()
        for preset_name, condition in self.preset.conditions:
            condition_str = str(condition)
            if condition_str not in already_evaluated:
                already_evaluated.add(condition_str)
                try:
                    self.check_condition(condition_str, context)
                except PresetAbortError as e:
                    raise PresetAbortError(f'Preset "{preset_name}" requested abort: {e} (--force to override)')

    @property
    def jinja_env(self):
        from jinja2.sandbox import SandboxedEnvironment

        global JINJA_ENV
        if JINJA_ENV is None:
            JINJA_ENV = SandboxedEnvironment()
        return JINJA_ENV

    def check_condition(self, condition_str, context):
        log.debug(f'Evaluating condition "{repr(condition_str)}"')
        template = self.jinja_env.from_string(condition_str)
        template.render(context)



from .base import ModuleTestBase, tempwordlist
from werkzeug.wrappers import Response
import re

from .test_module_paramminer_getparams import TestParamminer_Getparams
from .test_module_paramminer_headers import helper


class TestReflected_parameters_fromexcavate(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "reflected_parameters", "excavate"]

    def request_handler(self, request):
        normal_block = '<html><a href="/?reflected=foo">foo</a></html>'
        qs = str(request.query_string.decode())
        if "reflected=" in qs:
            value = qs.split("=")[1]
            if "&" in value:
                value = value.split("&")[0]
            reflected_block = f'<html><a href="/?reflected={value}"></a></html>'
            return Response(reflected_block, status=200)
        else:
            return Response(normal_block, status=200)

    async def setup_after_prep(self, module_test):
        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "[GETPARAM] Parameter value reflected in response body. Name: [reflected] Source Module: [excavate] Original Value: [foo]"
            for e in events
        )


class TestReflected_parameters_headers(TestReflected_parameters_fromexcavate):
    modules_overrides = ["httpx", "reflected_parameters", "excavate", "paramminer_headers"]
    config_overrides = {
        "modules": {
            "paramminer_headers": {"wordlist": tempwordlist(["junkword1", "tracestate"]), "recycle_words": True}
        }
    }

    def request_handler(self, request):
        headers = {k.lower(): v for k, v in request.headers.items()}
        if "tracestate" in headers:
            reflected_value = headers["tracestate"]
            reflected_block = f"<html><div>{reflected_value}</div></html>"
            return Response(reflected_block, status=200)
        else:
            return Response("<html><div></div></html>", status=200)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "[HEADER] Parameter value reflected in response body. Name: [tracestate] Source Module: [paramminer_headers]"
            for e in events
        )


class TestReflected_parameters_fromparamminer(TestParamminer_Getparams):
    modules_overrides = ["httpx", "paramminer_getparams", "reflected_parameters"]

    def request_handler(self, request):
        normal_block = "<html></html>"
        qs = str(request.query_string.decode())
        if "id=" in qs:
            value = qs.split("=")[1]
            if "&" in value:
                value = value.split("&")[0]
            reflected_block = f'<html><a href="/?id={value}"></a></html>'
            return Response(reflected_block, status=200)
        else:
            return Response(normal_block, status=200)

    async def setup_after_prep(self, module_test):
        module_test.scan.modules["paramminer_getparams"].rand_string = lambda *args, **kwargs: "AAAAAAAAAAAAAA"
        module_test.monkeypatch.setattr(
            helper.HttpCompare, "gen_cache_buster", lambda *args, **kwargs: {"AAAAAA": "1"}
        )

        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and "[GETPARAM] Parameter value reflected in response body. Name: [id] Source Module: [paramminer_getparams]"
            in e.data["description"]
            for e in events
        )


class TestReflected_parameters_with_canary(TestReflected_parameters_fromexcavate):
    def request_handler(self, request):
        normal_block = '<html><a href="/?reflected=foo">foo</a></html>'
        qs = str(request.query_string.decode())
        if qs:
            # Split the query string into key-value pairs
            params = qs.split("&")
            # Construct the reflected block with all parameters
            reflected_block = '<html><a href="/?'
            reflected_block += "&".join(params)
            reflected_block += '"></a></html>'
            return Response(reflected_block, status=200)
        else:
            return Response(normal_block, status=200)

    def check(self, module_test, events):
        # Ensure no findings are emitted when the canary is reflected
        assert not any(e.type == "FINDING" for e in events)


class TestReflected_parameters_cookies(TestReflected_parameters_fromexcavate):
    modules_overrides = ["httpx", "reflected_parameters", "excavate", "paramminer_cookies"]
    config_overrides = {
        "modules": {
            "paramminer_cookies": {"wordlist": tempwordlist(["junkword1", "testcookie"]), "recycle_words": True}
        }
    }

    def request_handler(self, request):
        cookies = request.cookies
        if "testcookie" in cookies:
            reflected_value = cookies["testcookie"]
            reflected_block = f"<html><div>{reflected_value}</div></html>"
            return Response(reflected_block, status=200)
        else:
            return Response("<html><div></div></html>", status=200)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "[COOKIE] Parameter value reflected in response body. Name: [testcookie] Source Module: [paramminer_cookies]"
            for e in events
        )


class TestReflected_parameters_postparams(TestReflected_parameters_fromexcavate):
    modules_overrides = ["httpx", "reflected_parameters", "excavate"]

    def request_handler(self, request):
        form_data = request.form
        if "testparam" in form_data:
            reflected_value = form_data["testparam"]
            reflected_block = f"<html><div>{reflected_value}</div></html>"
            return Response(reflected_block, status=200)
        else:
            form_html = """
            <html>
                <body>
                    <form action="/" method="post">
                        <input type="text" name="testparam" value="default_value">
                        <input type="submit" value="Submit">
                    </form>
                </body>
            </html>
            """
            return Response(form_html, status=200)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "[POSTPARAM] Parameter value reflected in response body. Name: [testparam] Source Module: [excavate] Original Value: [default_value]"
            for e in events
        )


class TestReflected_parameters_bodyjson(TestReflected_parameters_fromexcavate):
    modules_overrides = ["httpx", "reflected_parameters", "excavate"]

    def request_handler(self, request):
        # Ensure the request is expecting JSON data
        if request.content_type == "application/json":
            json_data = request.json
            if "username" in json_data:
                reflected_value = json_data["username"]
                reflected_block = f"<html><div>{reflected_value}</div></html>"
                return Response(reflected_block, status=200)
        # Provide an HTML page with a jQuery AJAX call
        jsonajax_extract_html = """
        <html>
        <script>
        function doLogin(e) {
          e.preventDefault();
          var username = $("#usernamefield").val();
          var password = $("#passwordfield").val();
          $.ajax({
            url: '/api/auth',
            type: 'POST',
            contentType: 'application/json',
            data: JSON.stringify({ username: username, password: password }),
            success: function (r) {
              window.location.replace("/demo");
            },
            error: function (r) {
              if (r.status == 401) {
                notify("Access denied");
              } else {
                notify(r.responseText);
              }
            }
          });
        }
        </script>
        <form action=/ method=GET><input type=text name="novalue"><button type=submit class=button>Submit</button></form>
        </html>
        """
        return Response(jsonajax_extract_html, status=200)

    async def setup_after_prep(self, module_test):
        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING"
            and e.data["description"]
            == "[BODYJSON] Parameter value reflected in response body. Name: [username] Source Module: [excavate]"
            for e in events
        )



import pytest
import string
import random
from bbot.scanner import Scanner


class TestBloomFilterBenchmarks:
    """
    Benchmark tests for Bloom Filter operations.

    These tests measure the performance of bloom filter operations which are
    critical for DNS brute-forcing efficiency in BBOT.
    """

    def setup_method(self):
        """Setup common test data"""
        self.scan = Scanner()

        # Generate test data of different sizes
        self.items_small = self._generate_random_strings(1000)  # 1K items
        self.items_medium = self._generate_random_strings(10000)  # 10K items

    def _generate_random_strings(self, n, length=10):
        """Generate a list of n random strings."""
        # Slightly longer strings for testing performance difference
        length = length + 2  # Make strings 2 chars longer
        return ["".join(random.choices(string.ascii_letters + string.digits, k=length)) for _ in range(n)]

    @pytest.mark.benchmark(group="bloom_filter_operations")
    def test_bloom_filter_dns_mutation_tracking_performance(self, benchmark):
        """Benchmark comprehensive bloom filter operations (add, check, mixed) for DNS brute-forcing"""

        def comprehensive_bloom_operations():
            bloom_filter = self.scan.helpers.bloom_filter(size=8000000)  # 8M bits

            # Phase 1: Add operations (simulating storing tried DNS mutations)
            for item in self.items_small:
                bloom_filter.add(item)

            # Phase 2: Check operations (simulating lookup of existing mutations)
            found_count = 0
            for item in self.items_small:
                if item in bloom_filter:
                    found_count += 1

            # Phase 3: Mixed operations (realistic DNS brute-force simulation)
            # Add new items while checking existing ones
            for i, item in enumerate(self.items_medium[:500]):  # Smaller subset for mixed ops
                bloom_filter.add(item)
                # Every few additions, check some existing items
                if i % 10 == 0:
                    for check_item in self.items_small[i : i + 5]:
                        if check_item in bloom_filter:
                            found_count += 1

            return {
                "items_added": len(self.items_small) + 500,
                "items_checked": found_count,
                "bloom_size": bloom_filter.size,
            }

        result = benchmark(comprehensive_bloom_operations)
        assert result["items_added"] > 1000
        assert result["items_checked"] > 0

    @pytest.mark.benchmark(group="bloom_filter_scalability")
    def test_bloom_filter_large_scale_dns_brute_force(self, benchmark):
        """Benchmark bloom filter performance with large-scale DNS brute-force simulation"""

        def large_scale_simulation():
            bloom_filter = self.scan.helpers.bloom_filter(size=8000000)  # 8M bits

            # Simulate a large DNS brute-force session
            mutations_tried = 0
            duplicate_attempts = 0

            # Add all medium dataset (simulating 10K DNS mutations)
            for item in self.items_medium:
                bloom_filter.add(item)
                mutations_tried += 1

            # Simulate checking for duplicates during brute-force
            for item in self.items_medium[:2000]:  # Check subset for duplicates
                if item in bloom_filter:
                    duplicate_attempts += 1

            # Simulate adding more mutations with duplicate checking
            for item in self.items_small:
                if item not in bloom_filter:  # Only add if not already tried
                    bloom_filter.add(item)
                    mutations_tried += 1
                else:
                    duplicate_attempts += 1

            return {
                "total_mutations_tried": mutations_tried,
                "duplicates_avoided": duplicate_attempts,
                "efficiency_ratio": mutations_tried / (mutations_tried + duplicate_attempts)
                if duplicate_attempts > 0
                else 1.0,
            }

        result = benchmark(large_scale_simulation)
        assert result["total_mutations_tried"] > 10000
        assert result["efficiency_ratio"] > 0



from .base import ModuleTestBase


class TestRetireJS(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "excavate", "retirejs"]

    # HTML page with vulnerable JavaScript libraries
    vulnerable_html = """<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8" />
  <title>retire.js test page</title>
</head>
<body>
  <h1>retire.js test page</h1>
  <p>This page includes JavaScript libraries for testing.</p>

  <!-- jQuery 3.4.1 -->
  <script src="/jquery-3.4.1.min.js"></script>

  <!-- Lodash 4.17.11 -->
  <script src="/lodash.min.js"></script>

  <!-- Handlebars 4.0.5 -->
  <script src="/handlebars.min.js"></script>

  <script>
    console.log('Libraries loaded');
  </script>
</body>
</html>"""

    # Sample jQuery 3.4.1 content
    jquery_content = """/*!
 * jQuery JavaScript Library v3.4.1
 * https://jquery.com/
 */
(function( global, factory ) {
    "use strict";
    factory( global );
})(typeof window !== "undefined" ? window : this, function( window, noGlobal ) {
    var jQuery = function( selector, context ) {
        return new jQuery.fn.init( selector, context );
    };
    jQuery.fn = jQuery.prototype = {};
    jQuery.fn.jquery = "3.4.1";
    if ( typeof noGlobal === "undefined" ) {
        window.jQuery = window.$ = jQuery;
    }
    return jQuery;
});"""

    # Sample Lodash 4.17.11 content
    lodash_content = """/**
 * @license
 * Lodash lodash.com/license | Underscore.js 1.8.3 underscorejs.org/LICENSE
 */
;(function(){
var i="4.17.11";
var Mn={VERSION:i};
if(typeof define=="function"&&define.amd)define(function(){return Mn});else if(typeof module=="object"&&module.exports)module.exports=Mn;else this._=Mn}());"""

    # Sample Handlebars 4.0.5 content
    handlebars_content = """/*!
 handlebars v4.0.5
*/
!function(a,b){"object"==typeof exports&&"object"==typeof module?module.exports=b():"function"==typeof define&&define.amd?define([],b):"object"==typeof exports?exports.Handlebars=b():a.Handlebars=b()}(this,function(){
var Handlebars={};
Handlebars.VERSION="4.0.5";
return Handlebars;
});"""

    async def setup_after_prep(self, module_test):
        expect_args = {"uri": "/"}
        respond_args = {"response_data": self.vulnerable_html}
        module_test.set_expect_requests(expect_args, respond_args)

        expect_args = {"uri": "/jquery-3.4.1.min.js"}
        respond_args = {"response_data": self.jquery_content}
        module_test.set_expect_requests(expect_args, respond_args)

        expect_args = {"uri": "/lodash.min.js"}
        respond_args = {"response_data": self.lodash_content}
        module_test.set_expect_requests(expect_args, respond_args)

        expect_args = {"uri": "/handlebars.min.js"}
        respond_args = {"response_data": self.handlebars_content}
        module_test.set_expect_requests(expect_args, respond_args)

    def check(self, module_test, events):
        # Check that excavate found the JavaScript URLs
        url_unverified_events = [e for e in events if e.type == "URL_UNVERIFIED"]
        js_url_events = [e for e in url_unverified_events if "extension-js" in e.tags]

        # Two out of the three URLs should be in the output
        # The third, non-vulnerable URL (lodash.min.js) is not output because it's a "special URL", and
        # nothing interesting has been discovered from it.
        vuln_urls = {"http://127.0.0.1:8888/handlebars.min.js", "http://127.0.0.1:8888/jquery-3.4.1.min.js"}
        assert {e.data for e in js_url_events} == vuln_urls, "Expected to find the vulnerable URLs in the output"

        # Check for FINDING events generated by retirejs
        finding_events = [e for e in events if e.type == "FINDING"]
        retirejs_findings = [
            e
            for e in finding_events
            if "vulnerable javascript library detected:" in e.data.get("description", "").lower()
        ]

        # We should have at least some findings from our vulnerable libraries
        assert len(retirejs_findings) > 0, (
            f"Expected retirejs to find vulnerabilities, but got {len(retirejs_findings)} findings"
        )

        # Check for specific expected vulnerability descriptions
        descriptions = [finding.data.get("description", "") for finding in retirejs_findings]
        all_descriptions = "\n".join(descriptions)

        # Look for specific vulnerabilities we expect to find
        expected_handlebars_vuln = "Vulnerable JavaScript library detected: handlebars v4.0.5 Severity: HIGH Summary: Regular Expression Denial of Service in Handlebars JavaScript URL: http://127.0.0.1:8888/handlebars.min.js CVE(s): CVE-2019-20922 Affected versions: [4.0.0 to 4.4.5)"
        expected_jquery_vuln = "Vulnerable JavaScript library detected: jquery v3.4.1 Severity: MEDIUM Summary: Regex in its jQuery.htmlPrefilter sometimes may introduce XSS JavaScript URL: http://127.0.0.1:8888/jquery-3.4.1.min.js CVE(s): CVE-2020-11022 Affected versions: [1.2.0 to 3.5.0)"

        # Verify at least one of the expected vulnerabilities is found
        handlebars_found = expected_handlebars_vuln in all_descriptions
        jquery_found = expected_jquery_vuln in all_descriptions

        assert handlebars_found and jquery_found, (
            f"Expected to find specific vulnerabilities but didn't find them. Found descriptions:\n{all_descriptions}"
        )

        # Basic validation of findings structure
        for finding in retirejs_findings:
            assert "description" in finding.data, "Finding should have description"
            assert "url" in finding.data, "Finding should have url"
            assert finding.parent.type == "URL_UNVERIFIED", "Parent should be URL_UNVERIFIED"


class TestRetireJSNoExcavate(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "retirejs"]
    force_start = True  # Allow scan to continue even if modules fail setup
    config_overrides = {
        "excavate": False,
    }

    def check(self, module_test, events):
        # When excavate is disabled, retirejs should fail setup but scan should still run
        retirejs_module = module_test.scan.modules.get("retirejs")

        if retirejs_module:
            # Check that the module exists but setup failed
            setup_status = getattr(retirejs_module, "_setup_status", None)
            if setup_status is not None:
                success, error_msg = setup_status
                assert success is False, "retirejs setup should have failed without excavate"
                expected_error = "retirejs will not function without excavate enabled"
                assert error_msg == expected_error, f"Expected error message '{expected_error}', but got '{error_msg}'"

        # No retirejs findings should be generated since setup failed
        retirejs_findings = [e for e in events if e.type == "FINDING" and getattr(e, "module", None) == "retirejs"]
        assert len(retirejs_findings) == 0, "retirejs should not generate findings when setup fails"



import json
from pathlib import Path
from bbot.errors import WordlistError
from bbot.modules.base import BaseModule

# key: <common-protocol-name> value: <legba-protocol-plugin-name>
# List with `legba -L`
PROTOCOL_LEGBA_PLUGIN_MAP = {
    "postgresql": "pgsql",
}


# Maps common protocol names to Legba protocol plugin names
def map_protocol_to_legba_plugin_name(common_protocol_name: str) -> str:
    return PROTOCOL_LEGBA_PLUGIN_MAP.get(common_protocol_name, common_protocol_name)


class legba(BaseModule):
    watched_events = ["PROTOCOL"]
    produced_events = ["FINDING"]
    flags = ["active", "aggressive", "deadly"]
    per_hostport_only = True
    meta = {
        "description": "Credential bruteforcing supporting various services.",
        "created_date": "2025-07-18",
        "author": "@christianfl, @fuzikowski",
    }
    _module_threads = 25
    scope_distance_modifier = None

    options = {
        "ssh_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/ssh-betterdefaultpasslist.txt",
        "ftp_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/ftp-betterdefaultpasslist.txt",
        "telnet_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/telnet-betterdefaultpasslist.txt",
        "vnc_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/vnc-betterdefaultpasslist.txt",
        "mssql_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/mssql-betterdefaultpasslist.txt",
        "mysql_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/mysql-betterdefaultpasslist.txt",
        "postgresql_wordlist": "https://raw.githubusercontent.com/danielmiessler/SecLists/refs/heads/master/Passwords/Default-Credentials/postgres-betterdefaultpasslist.txt",
        "concurrency": 3,
        "rate_limit": 3,
        "version": "1.1.1",
    }

    options_desc = {
        "ssh_wordlist": "Wordlist URL for SSH combined username:password wordlist, newline separated",
        "ftp_wordlist": "Wordlist URL for FTP combined username:password wordlist, newline separated",
        "telnet_wordlist": "Wordlist URL for TELNET combined username:password wordlist, newline separated",
        "vnc_wordlist": "Wordlist URL for VNC password wordlist, newline separated",
        "mssql_wordlist": "Wordlist URL for MSSQL combined username:password wordlist, newline separated",
        "mysql_wordlist": "Wordlist URL for MySQL combined username:password wordlist, newline separated",
        "postgresql_wordlist": "Wordlist URL for PostgreSQL combined username:password wordlist, newline separated",
        "concurrency": "Number of concurrent workers, gets overridden for SSH",
        "rate_limit": "Limit the number of requests per second, gets overridden for SSH",
        "version": "legba version",
    }

    deps_ansible = [
        {
            "name": "Download legba",
            "unarchive": {
                "src": "https://github.com/evilsocket/legba/releases/download/#{BBOT_MODULES_LEGBA_VERSION}/legba-#{BBOT_MODULES_LEGBA_VERSION}-#{BBOT_OS}-#{BBOT_CPU_ARCH_RUST}.tar.gz",
                "dest": "#{BBOT_TEMP}",
                "include": "legba-#{BBOT_MODULES_LEGBA_VERSION}-#{BBOT_OS}-#{BBOT_CPU_ARCH_RUST}/legba",
                "remote_src": True,
                "mode": "u+x,g+x,o+x",
            },
        }
    ]

    async def setup(self):
        self.output_dir = Path(self.scan.temp_dir / "legba-output")
        self.helpers.mkdir(self.output_dir)
        if not "fingerprintx" in self.scan.modules:
            self.warning("Enabling 'fingerprintx' module is recommended for discovery of PROTOCOL events")

        return True

    async def filter_event(self, event):
        handled_protocols = ["ssh", "ftp", "telnet", "vnc", "mssql", "mysql", "postgresql"]

        protocol = event.data["protocol"].lower()
        if not protocol in handled_protocols:
            return False, f"service {protocol} is currently not supported or can't be bruteforced by Legba"

        return True

    async def handle_event(self, event):
        host = str(event.host)
        port = str(event.port)
        protocol = event.data["protocol"].lower()

        command_data = await self.construct_command(host, port, protocol)

        if not command_data:
            self.warning(f"Skipping {host}:{port} ({protocol}) due to errors while constructing the command")
            return

        command, output_path = command_data

        await self.run_process(command)

        async for finding_event in self.parse_output(output_path, event):
            await self.emit_event(finding_event)

    async def parse_output(self, output_filepath, event):
        protocol = event.data["protocol"].lower()

        try:
            with open(output_filepath) as file:
                for line in file:
                    # example line (ssh):
                    # {"found_at":"2025-07-18T06:28:08.969812152+01:00","target":"localhost:22","plugin":"ssh","data":{"username":"user","password":"pass"},"partial":false}
                    line = line.strip()

                    try:
                        data = json.loads(line)["data"]
                        username = data.get("username", "")
                        password = data.get("password", "")

                        if username and password:
                            message_addition = f"{username}:{password}"
                        elif username:
                            message_addition = username
                        elif password:
                            message_addition = password
                    except Exception as e:
                        self.warning(f"Failed to parse Legba output ({line}), using raw output instead: {e}")
                        message_addition = f"raw output: {line}"

                    yield self.make_event(
                        {
                            "severity": "CRITICAL",
                            "confidence": "CONFIRMED",
                            "host": str(event.host),
                            "port": str(event.port),
                            "description": f"Valid {protocol} credentials found - {message_addition}",
                        },
                        "FINDING",
                        parent=event,
                    )
        except FileNotFoundError:
            self.info(
                f"Could not open Legba output file {output_filepath}. File is missing if no valid credentials could be found"
            )
        except Exception as e:
            self.warning(f"Error processing Legba output file {output_filepath}: {e}")
        else:
            self.helpers.delete_file(output_filepath)

    async def construct_command(self, host, port, protocol):
        # -C                Combo wordlist delimited by ':'
        # -P                Passwordlist
        # --target          Target (allowed: host, url, IP address, CIDR, @filename)
        # --output-format   Output file format
        # --output          Save results to this file
        # -Q                Do not report statistics
        #
        # --wait            Wait time in milliseconds per login attempt
        # --rate-limit      Limit the number of requests per second
        # --concurrency     Number of concurrent workers

        # Example command to bruteforce SSH:
        #
        # legba ssh -C combolist.txt --target 127.0.0.1:22 --output-format jsonl --output out.txt -Q --wait 4000 --rate-limit 1 --concurrency 1

        try:
            wordlist_path = await self.helpers.wordlist(self.config.get(f"{protocol}_wordlist"))
        except WordlistError as e:
            self.warning(f"Error retrieving wordlist for protocol {protocol}: {e}")
            return None
        except Exception as e:
            self.warning(f"Unexpected error during wordlist loading for protocol {protocol}: {e}")
            return None

        protocol_plugin_name = map_protocol_to_legba_plugin_name(protocol)
        output_path = Path(self.output_dir) / f"{host}_{port}.json"

        cmd = [
            "legba",
            protocol_plugin_name,
        ]

        if protocol == "vnc":
            # use only passwords, not combinations
            cmd += ["-P"]

        else:
            # use combinations
            cmd += ["-C"]

        # wrap IPv6 addresses in square brackets
        if self.helpers.is_ip(host, version=6):
            host = f"[{host}]"

        cmd += [
            wordlist_path,
            "--target",
            f"{host}:{port}",
            "--output-format",
            "jsonl",
            "--output",
            output_path,
            "-Q",
        ]

        if protocol == "ssh":
            # With OpenSSH 9.8, the sshd_config option "PerSourcePenalties" was introduced (on by default)
            # The penalty "authfail" defaults to 5 seconds, so bruteforcing fast will block access.
            # Legba is not able to check that by itself, so the wait time is set to 5 s, rate limit to 1 and concurrency to 1 with SSH.
            # See https://www.openssh.com/txt/release-9.8
            cmd += [
                "--wait",
                "5000",
                "--rate-limit",
                "1",
                "--concurrency",
                "1",
            ]
        else:
            cmd += ["--rate-limit", self.config.rate_limit, "--concurrency", self.config.concurrency]

        return cmd, output_path



import copy
import base64
import binascii
from urllib.parse import quote


class BaseLightfuzz:
    friendly_name = ""
    uses_interactsh = False

    def __init__(self, lightfuzz, event):
        self.lightfuzz = lightfuzz
        self.event = event
        self.results = []
        self.parameter_name = self.event.data["name"]

    @staticmethod
    def is_hex(s):
        try:
            bytes.fromhex(s)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_base64(s):
        try:
            if base64.b64encode(base64.b64decode(s)).decode() == s:
                return True
        except (binascii.Error, UnicodeDecodeError):
            return False
        return False

    # WEB_PARAMETER event may contain additional_params (e.g. other parameters in the same form or query string). These will be sent unchanged along with the probe.
    def additional_params_process(self, additional_params, additional_params_populate_empty):
        """
        Processes additional parameters by populating blank or empty values with random strings if specified.

        Parameters:
        - additional_params (dict): A dictionary of additional parameters to process.
        - additional_params_populate_blank_empty (bool): If True, populates blank or empty parameter values with random numeric strings.

        Returns:
        - dict: A dictionary with processed additional parameters, where blank or empty values are replaced with random strings if specified.

        The function iterates over the provided additional parameters and replaces any blank or empty values with a random numeric string
        of length 10, if the flag is set to True. Otherwise, it returns the parameters unchanged.
        """
        if not additional_params or not additional_params_populate_empty:
            return additional_params

        return {
            k: self.lightfuzz.helpers.rand_string(10, numeric_only=True) if v in ("", None) else v
            for k, v in additional_params.items()
        }

    def conditional_urlencode(self, probe, event_type, skip_urlencoding=False):
        """Conditionally url-encodes the probe if the event type requires it and encoding is not skipped by the submodule.
        We also don't encode if any envelopes are present.
        """
        if event_type in ["GETPARAM", "COOKIE"] and not skip_urlencoding and getattr(self.event, "envelopes", None):
            # Exclude '&' from being encoded since we are operating on full query strings
            return quote(probe, safe="&")
        return probe

    def build_query_string(self, probe, parameter_name, additional_params=None):
        """Constructs a URL with query parameters from the given probe and additional parameters."""
        url = f"{self.event.data['url']}?{parameter_name}={probe}"
        if additional_params:
            url = self.lightfuzz.helpers.add_get_params(url, additional_params, encode=False).geturl()
        return url

    def prepare_request(
        self,
        event_type,
        probe,
        cookies,
        additional_params=None,
        speculative_mode="GETPARAM",
        parameter_name_suffix="",
        additional_params_populate_empty=False,
        skip_urlencoding=False,
    ):
        """
        Prepares the request parameters by processing the probe and constructing the request based on the event type.
        """

        if parameter_name_suffix:
            parameter_name = f"{self.parameter_name}{parameter_name_suffix}"
        else:
            parameter_name = self.parameter_name
        additional_params = self.additional_params_process(additional_params, additional_params_populate_empty)

        # Transparently pack the probe value into the envelopes, if present
        probe = self.outgoing_probe_value(probe)

        # URL Encode the probe if the event type is GETPARAM or COOKIE, if there are no envelopes, and the submodule did not opt-out with skip_urlencoding
        probe = self.conditional_urlencode(probe, event_type, skip_urlencoding)

        if event_type == "SPECULATIVE":
            event_type = speculative_mode

        # Construct request parameters based on the event type
        if event_type == "GETPARAM":
            url = self.build_query_string(probe, parameter_name, additional_params)
            return {"method": "GET", "cookies": cookies, "url": url}
        elif event_type == "COOKIE":
            cookies_probe = {parameter_name: probe}
            return {"method": "GET", "cookies": {**cookies, **cookies_probe}, "url": self.event.data["url"]}
        elif event_type == "HEADER":
            headers = {parameter_name: probe}
            return {"method": "GET", "headers": headers, "cookies": cookies, "url": self.event.data["url"]}
        elif event_type in ["POSTPARAM", "BODYJSON"]:
            # Prepare data for POSTPARAM and BODYJSON event types
            data = {parameter_name: probe}
            if additional_params:
                data.update(additional_params)
            if event_type == "BODYJSON":
                return {"method": "POST", "json": data, "cookies": cookies, "url": self.event.data["url"]}
            else:
                return {"method": "POST", "data": data, "cookies": cookies, "url": self.event.data["url"]}

    def compare_baseline(
        self,
        event_type,
        probe,
        cookies,
        additional_params_populate_empty=False,
        speculative_mode="GETPARAM",
        skip_urlencoding=False,
        parameter_name_suffix="",
        parameter_name_suffix_additional_params="",
    ):
        """
        Compares the baseline using prepared request parameters.
        """
        additional_params = copy.deepcopy(self.event.data.get("additional_params", {}))

        if additional_params and parameter_name_suffix_additional_params:
            # Add suffix to each key in additional_params
            additional_params = {
                f"{k}{parameter_name_suffix_additional_params}": v for k, v in additional_params.items()
            }

        request_params = self.prepare_request(
            event_type,
            probe,
            cookies,
            additional_params,
            speculative_mode,
            parameter_name_suffix,
            additional_params_populate_empty,
            skip_urlencoding,
        )
        request_params.update({"include_cache_buster": False})
        return self.lightfuzz.helpers.http_compare(**request_params)

    async def baseline_probe(self, cookies):
        """
        Executes a baseline probe to establish a baseline for comparison.
        """
        if self.event.data.get("eventtype") in ["POSTPARAM", "BODYJSON"]:
            method = "POST"
        else:
            method = "GET"

        return await self.lightfuzz.helpers.request(
            method=method,
            cookies=cookies,
            url=self.event.data.get("url"),
            allow_redirects=False,
            retries=1,
            timeout=10,
        )

    async def compare_probe(
        self,
        http_compare,
        event_type,
        probe,
        cookies,
        additional_params_populate_empty=False,
        additional_params_override={},
        speculative_mode="GETPARAM",
        skip_urlencoding=False,
        parameter_name_suffix="",
        parameter_name_suffix_additional_params="",
    ):
        # Deep copy to avoid modifying original additional_params
        additional_params = copy.deepcopy(self.event.data.get("additional_params", {}))

        # Override additional parameters if provided
        additional_params.update(additional_params_override)

        if additional_params and parameter_name_suffix_additional_params:
            # Add suffix to each key in additional_params
            additional_params = {
                f"{k}{parameter_name_suffix_additional_params}": v for k, v in additional_params.items()
            }

        # Prepare request parameters
        request_params = self.prepare_request(
            event_type,
            probe,
            cookies,
            additional_params,
            speculative_mode,
            parameter_name_suffix,
            additional_params_populate_empty,
            skip_urlencoding,
        )
        # Perform the comparison using the constructed request parameters
        url = request_params.pop("url")
        return await http_compare.compare(url, **request_params)

    async def standard_probe(
        self,
        event_type,
        cookies,
        probe,
        timeout=10,
        additional_params_populate_empty=False,
        speculative_mode="GETPARAM",
        allow_redirects=False,
        skip_urlencoding=False,
    ):
        request_params = self.prepare_request(
            event_type,
            probe,
            cookies,
            self.event.data.get("additional_params"),
            speculative_mode,
            "",
            additional_params_populate_empty,
            skip_urlencoding,
        )
        request_params.update({"allow_redirects": allow_redirects, "retries": 0, "timeout": timeout})
        self.debug(f"standard_probe requested URL: [{request_params['url']}]")
        return await self.lightfuzz.helpers.request(**request_params)

    def metadata(self):
        metadata_string = f"Parameter: [{self.event.data['name']}] Parameter Type: [{self.event.data['type']}]"
        if self.event.data["original_value"] != "" and self.event.data["original_value"] is not None:
            metadata_string += (
                f" Original Value: [{self.lightfuzz.helpers.truncate_string(self.event.data['original_value'], 200)}]"
            )
        return metadata_string

    def incoming_probe_value(self, populate_empty=True):
        """
        Transparently modifies the incoming probe value (the original value of the WEB_PARAMETER), given any envelopes that may have been identified, so that fuzzing within the envelopes can occur.
        """
        envelopes = getattr(self.event, "envelopes", None)
        probe_value = ""
        if envelopes is not None:
            probe_value = envelopes.get_subparam()
            self.debug(f"incoming_probe_value (after unpacking): {probe_value} with envelopes [{envelopes}]")
        if not probe_value:
            if populate_empty is True:
                probe_value = self.lightfuzz.helpers.rand_string(10, numeric_only=True)
            else:
                probe_value = ""
        probe_value = str(probe_value)
        return probe_value

    def outgoing_probe_value(self, outgoing_probe_value):
        """
        Transparently packs the outgoing probe value (fuzz probe being sent to the target) through
        any envelopes that may have been identified, so that fuzzing within the envelopes can occur.

        Uses pack_value() to avoid mutating the envelope's internal state, preventing cross-contamination
        between submodules that share the same event/envelope object.
        """
        self.debug(f"outgoing_probe_value (before packing): {outgoing_probe_value} / {self.event}")
        envelopes = getattr(self.event, "envelopes", None)
        if envelopes is not None:
            outgoing_probe_value = envelopes.pack_value(outgoing_probe_value)
            self.debug(
                f"outgoing_probe_value (after packing): {outgoing_probe_value} with envelopes [{envelopes}] / {self.event}"
            )
        return outgoing_probe_value

    def get_submodule_name(self):
        """Extracts the submodule name from the class name."""
        return self.__class__.__name__.replace("Lightfuzz", "").lower()

    def log(self, level, message, *args, **kwargs):
        submodule_name = self.get_submodule_name()
        prefixed_message = f"[{submodule_name}] {message}"
        log_method = getattr(self.lightfuzz, level)
        log_method(prefixed_message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        self.log("debug", message, *args, **kwargs)

    def verbose(self, message, *args, **kwargs):
        self.log("verbose", message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        self.log("info", message, *args, **kwargs)

    def hugeinfo(self, message, *args, **kwargs):
        self.log("hugeinfo", message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.log("warning", message, *args, **kwargs)

    def hugewarning(self, message, *args, **kwargs):
        self.log("hugewarning", message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.log("error", message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        self.log("critical", message, *args, **kwargs)



import yara


class YaraHelper:
    def __init__(self, parent_helper):
        self.parent_helper = parent_helper

    def compile_strings(self, strings: list[str], nocase=False):
        """
        Compile a list of strings into a YARA rule
        """
        # Format each string as a YARA string definition
        yara_strings = []
        for i, s in enumerate(strings):
            s = s.replace('"', '\\"')
            yara_string = f'$s{i} = "{s}"'
            if nocase:
                yara_string += " nocase"
            yara_strings.append(yara_string)
        yara_strings = "\n        ".join(yara_strings)

        # Create the complete YARA rule
        yara_rule = f"""
rule strings_match
{{
    strings:
        {yara_strings}
    condition:
        any of them
}}
"""
        # Compile and return the rule
        return self.compile(source=yara_rule)

    def compile(self, *args, **kwargs):
        return yara.compile(*args, **kwargs)

    async def match(self, compiled_rules, text):
        """
        Given a compiled YARA rule and a body of text, return a list of strings that match the rule
        """
        matched_strings = []
        matches = await self.parent_helper.run_in_executor(compiled_rules.match, data=text)
        if matches:
            for match in matches:
                for string_match in match.strings:
                    for instance in string_match.instances:
                        matched_string = instance.matched_data.decode("utf-8")
                        matched_strings.append(matched_string)
        return matched_strings
